from ansys.mapdl.core._commands import parse


def bsplin(self, p1="", p2="", p3="", p4="", p5="", p6="", xv1="", yv1="",
           zv1="", xv6="", yv6="", zv6="", **kwargs) -> int:
    """Generate a single line from a spline fit to a series of keypoints.

    APDL Command: BSPLIN

    Parameters
    ----------
    p1, p2, p3, p4, p5, p6
        Keypoints through which a spline is fit.  At least two
        keypoints must be defined.

    XV1, YV1, ZV1
        Orientation point of an outward vector tangent to line at
        P1. Vector coordinate system has its origin at the
        keypoint. Coordinate interpretation corresponds to the
        active coordinate system type, i.e., X is R for
        cylindrical, etc. Defaults to zero curvature slope.

    XV6, YV6, ZV6
        Orientation point of an outward vector tangent to a line
        at P6 (or the last keypoint specified if fewer than six
        specified). Defaults to zero curvature slope.

    Returns
    -------
    int
        Line number of the spline generated from the spline fit.

    Examples
    --------
    Generate a spline through ``(0, 0, 0)``, ``(0, 1, 0)`` and
    ``(1, 2, 0)``

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 1, 0)
    >>> k2 = mapdl.k("", 1, 2, 0)
    >>> lnum = mapdl.bsplin(k0, k1, k2)

    Notes
    -----
    One line is generated between keypoint P1 and the last
    keypoint entered.  The line will pass through each entered
    keypoint.  Solid modeling in a toroidal coordinate system is
    not recommended.
    """
    command = f"BSPLIN,{p1},{p2},{p3},{p4},{p5},{p6},{xv1},{yv1},{zv1},{xv6},{yv6},{zv6}"
    return parse.parse_line_no(self.run(command, **kwargs))


def circle(self, pcent="", rad="", paxis="", pzero="", arc="", nseg="",
           **kwargs) -> list:
    """Generate circular arc lines.

    APDL Command: CIRCLE

    Parameters
    ----------
    pcent
        Keypoint defining the center of the circle (in the plane
        of the circle).

    rad
        Radius of the circle.  If RAD is blank and PCENT = P, the
        radius is the distance from PCENT to PZERO.

    paxis
        Keypoint defining axis of circle (along with PCENT).  If
        PCENT = P and PAXIS is omitted, the axis is normal to the
        working plane.

    pzero
        Keypoint defining the plane normal to circle (along with
        PCENT and PAXIS) and the zero degree location.  Need not
        be in the plane of the circle. This value is not required
        if PAXIS is defined along the Y axis (that is, a circle in
        the XZ plane).

    arc
        Arc length (in degrees).  Positive follows right-hand rule
        about PCENT-PAXIS vector.  Defaults to 360 degrees.

    nseg
        Number of lines around circumference (defaults to minimum
        required for 90 degrees-maximum arcs, i.e., 4 for 360 degrees).  Number
        of keypoints generated is NSEG for 360 degrees or NSEG + 1 for
        less than 360 degrees.

    Returns
    -------
    list
        List of lines of the circular arcs generated from this
        command.

    Examples
    --------
    Create a full circle containing four circular arcs.  Circle
    centered at (0, 0, 0) and generated in the XY plane.  Return
    the lines generated from the circle.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 0, 1)
    >>> carc0 = mapdl.circle(k0, 1, k1)
    >>> carc0
    [1, 2, 3, 4]

    Notes
    -----
    Generates circular arc lines (and their corresponding
    keypoints).  Keypoints are generated at regular angular
    locations (based on a maximum spacing of 90 degrees).  Arc lines are
    generated connecting the keypoints.  Keypoint and line numbers
    are automatically assigned, beginning with the lowest
    available values [NUMSTR].  Adjacent lines use a common
    keypoint.  Line shapes are generated as arcs, regardless of
    the active coordinate system.  Line shapes are invariant with
    coordinate system after they are generated.
    """
    command = f"CIRCLE,{pcent},{rad},{paxis},{pzero},{arc},{nseg}"
    return parse.parse_line_nos(self.run(command, **kwargs))


def l(self, p1="", p2="", ndiv="", space="", xv1="", yv1="", zv1="",
      xv2="", yv2="", zv2="", **kwargs) -> int:
    """Define a line between two keypoints.

    APDL Command: L

    Parameters
    ----------
    p1
        Keypoint at the beginning of line.

    p2
        Keypoint at the end of line.

    ndiv
        Number of element divisions within this line.  Normally
        this field is not used; specifying divisions with LESIZE,
        etc. is recommended.

    space
        Spacing ratio.  Normally this field is not used, as
        specifying spacing ratios with the LESIZE command is
        recommended.  If positive, space is the nominal ratio of
        the last division size (at P2) to the first division size
        (at P1).  If the ratio is greater than 1, the division
        sizes increase from P1 to P2, and if less than 1, they
        decrease.  If space is negative, then ``space`` is the
        nominal ratio of the center division size to those at the
        ends.

    Returns
    -------
    int
        The line number of the generated line.

    Examples
    --------
    Create a line between the two keypoints (0, 0, 0) and (1, 0, 0)

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> lnum = mapdl.l(k0, k1)
    >>> lnum
    1

    Notes
    -----
    Defines a line between two keypoints from P1 to P2.  The line
    shape may be generated as "straight" (in the active coordinate
    system) or curved.  The line shape is invariant with
    coordinate system after it is generated.  Note that solid
    modeling in a toroidal coordinate system is not recommended.
    A curved line is limited to 180 degrees.  Lines may be redefined only
    if not yet attached to an area.
    """
    command = f"L,{p1},{p2},{ndiv},{space},{xv1},{yv1},{zv1},{xv2},{yv2},{zv2}"
    return parse.parse_line_no(self.run(command, **kwargs))


def l2ang(self, nl1="", nl2="", ang1="", ang2="", phit1="", phit2="",
          **kwargs) -> int:
    """Generates a line at an angle with two existing lines.

    APDL Command: L2ANG

    Parameters
    ----------
    nl1
        Number of the first line to be hit (touched by the end of
        the new line).  If negative, assume P1 (see below) is the
        second keypoint of the line instead of the first.

    nl2
        Number of the second line to be hit.  If negative, assume
        P3 is the second keypoint of the line instead of the
        first.

    ang1
        Angle of intersection (usually zero or 180) of generated
        line with tangent to first line.

    ang2
        Angle of intersection (usually zero or 180) of generated
        line with tangent to second line.

    phit1
        Number to be assigned to keypoint generated at hit
        location on first line (defaults to lowest available
        keypoint number [NUMSTR]).

    phit2
        Number to be assigned to keypoint generated at hit
        location on second line (defaults to lowest available
        keypoint number [NUMSTR]).

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create two circles and join them with a line.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 0, 1)
    >>> k2 = mapdl.k("", 0, 0, 0.5)
    >>> carc0 = mapdl.circle(k0, 1, k1)
    >>> carc1 = mapdl.circle(k2, 1, k1)
    >>> lnum = mapdl.l2ang(carc0[0], carc1[0], 90, 90)
    >>> lnum
    9

    Notes
    -----
    Generates a straight line (PHIT1-PHIT2) at an angle (ANG1)
    with an existing line NL1 (P1-P2) and which is also at an
    angle (ANG2) with another existing line NL2 (P3-P4).  If the
    angles are zero the generated line is tangent to the two
    lines.  The PHIT1 and PHIT2 locations on the lines are
    automatically calculated.  Line P1-P2 becomes P1-PHIT1, P3-P4
    becomes P3-PHIT2, and new lines PHIT1-P2, PHIT2-P4, and
    PHIT1-PHIT2 are generated.  Line divisions are set to zero
    (use LESIZE, etc. to modify).
    """
    command = f"L2ANG,{nl1},{nl2},{ang1},{ang2},{phit1},{phit2}"
    msg = self.run(command, **kwargs)
    if msg:
        return parse.parse_line_no(msg)


def l2tan(self, nl1="", nl2="", **kwargs) -> int:
    """Generates a line tangent to two lines.

    APDL Command: L2TAN

    Parameters
    ----------
    nl1
        Number of the first line generated line is tangent to.  If
        negative, assume P1 (see below) is the second keypoint of
        the line instead of the first.

    nl2
        Number of the second line generated line is tangent to.
        If negative, assume P3 is the second keypoint of the line
        instead of the first.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create two circular arcs and connect them with a spline.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 0, 1)
    >>> k2 = mapdl.k("", -1.5, 1.5, 0)
    >>> k3 = mapdl.k("", -1.5, 1.5, 1)
    >>> carc0 = mapdl.circle(k0, 1, k1, arc=90)
    >>> carc1 = mapdl.circle(k2, 1, k3, arc=90)
    >>> lnum = mapdl.l2tan(1, 2)
    3

    Plot these lines

    >>> mapdl.lplot(cpos='xy')

    Notes
    -----
    Generates a line (P2-P3) tangent at point P2 to line NL1
    (P1-P2) and tangent at point P3 to line NL2 (P3-P4).
    """
    command = f"L2TAN,{nl1},{nl2}"
    return parse.parse_line_no(self.run(command, **kwargs))


def lang(self, nl1="", p3="", ang="", phit="", locat="", **kwargs) -> int:
    """Generate a straight line at an angle with a line.

    APDL Command: LANG

    Parameters
    ----------
    nl1
        Number of the line to be hit (touched by the end of the
        new line).  If negative, assume P1 (see below) is the
        second keypoint of the line instead of the first.

    p3
        Keypoint at which generated line must end.

    ang
        Angle of intersection of generated line PHIT-P3 with
        tangent to line P1-P2 at PHIT.  If 0 (default), the
        generated line is tangent to NL1 toward end P1; if 90, the
        generated line is perpendicular to NL1.  If 180, the
        generated line is tangent to NL1 toward end P2.  ANG can
        be any value, but is adjusted to the corresponding acute
        angle with respect to LOCAT. See "Notes" for a discussion
        of accuracy.

    phit
        Number to be assigned to keypoint generated at hit
        location (defaults to lowest available keypoint number
        [NUMSTR]).

    locat
        Approximate location of PHIT in terms of the ratio of the
        distance along the line (NL1) to the length of the line.
        LOCAT can range from 0 to 1.  If LOCAT is blank, the point
        will be located with less speed and accuracy, and an
        arbitrary location may result.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create a line from a line from (0, 0, 0) to (1, 0, 0) to a
    keypoint at (1, 1, 1) at an angle of 60 degrees.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> lnum = mapdl.l(k0, k1)
    >>> k2 = mapdl.k("", 1, 1, 0)
    >>> lnum = mapdl.lang(lnum, k2, 60)
    >>> lnum
    2

    Notes
    -----
    Generates a straight line (PHIT-P3) at an angle (ANG) with a
    line NL1 (P1-P2).  The location of PHIT on the line is
    automatically calculated.  Line P1-P2 becomes P1-PHIT and new
    lines PHIT-P2 and PHIT-P3 are generated.  Line divisions are
    set to zero (use LESIZE, etc. to modify).

    PHIT is positioned closest to LOCAT for the given angle, ANG.
    To ensure better performance, it is recommended that LOCAT be
    input, even if it is 0.

    The program uses an iterative procedure to position PHIT.  The
    procedure is not exact, with the result that the actual value
    of ANG will sometimes differ slightly from the specified
    value.
    """
    command = f"LANG,{nl1},{p3},{ang},{phit},{locat}"
    return parse.parse_line_no(self.run(command, **kwargs))


def larc(self, p1="", p2="", pc="", rad="", **kwargs) -> int:
    """Define a circular arc.

    APDL Command: LARC

    Parameters
    ----------
    p1
        Keypoint at one end of circular arc line.

    p2
        Keypoint at other end of circular arc line.

    pc
        Keypoint defining plane of arc and center of curvature
        side (with positive radius).  Must not lie along the
        straight line from P1 to P2.  PC need not be at the center
        of curvature.

    rad
        Radius of curvature of the arc.  If negative, assume
        center of curvature side is opposite to that defined by
        PC.  If RAD is blank, RAD will be calculated from a curve
        fit through P1, PC, and P2.

    Returns
    -------
    int
        Line number of the arc.

    Examples
    --------
    Create a circular arc that travels between (0, 0, 0) and
    (1, 1, 0) with a radius of curvature of 2.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 1, 0)
    >>> k2 = mapdl.k("", 0, 1, 0)
    >>> lnum = mapdl.larc(k0, k1, k2, 2)
    1

    Notes
    -----
    Defines a circular arc line from P1 to P2.  The line shape is
    generated as circular, regardless of the active coordinate
    system.  The line shape is invariant with coordinate system
    after it is generated.

    When dealing with a large radius arc (1e3), or if the location
    of the arc you create is far away from the origin of your
    coordinate system, anomalies may occur. You can prevent this
    by creating the arc at a smaller scale, and then scaling the
    model back to full size (LSSCALE).
    """
    command = f"LARC,{p1},{p2},{pc},{rad}"
    return parse.parse_line_no(self.run(command, **kwargs))


def larea(self, p1="", p2="", narea="", **kwargs) -> int:
    """Generate the shortest line between two keypoints on an area.

    APDL Command: LAREA

    Parameters
    ----------
    p1
        First keypoint of line to be generated.

    p2
        Second keypoint of line to be generated.

    narea
        Area containing P1 and P2, or area to which generated line
        is to be parallel.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Generate a line on a square between its two corners.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> k2 = mapdl.k("", 1, 1, 0)
    >>> k3 = mapdl.k("", 0, 1, 0)
    >>> a0 = mapdl.a(k0, k1, k2, k3)
    >>> lnum = mapdl.larea(k0, k2, a0)
    >>> lnum
    1

    Notes
    -----
    Generates the shortest line between two keypoints, P1 and P2,
    both of which lie on an area.  The generated line will also
    lie on the area.  P1 and P2 may also be equidistant (in global
    Cartesian space) from the area (and on the same side of the
    area), in which case a line parallel to the area is generated.
    """
    command = f"LAREA,{p1},{p2},{narea}"
    return parse.parse_line_no(self.run(command, **kwargs))


def lcomb(self, nl1="", nl2="", keep="", **kwargs) -> int:
    """Combines adjacent lines into one line.

    APDL Command: LCOMB

    Parameters
    ----------
    nl1
        Number of the first line to be combined.  If NL1 = ALL,
        NL2 is ignored and all selected lines [LSEL] are combined.
        A component name may also be substituted for NL1 (NL2 is
        ignored).

    nl2
        Number of the second line to be combined.

    keep
        Specifies whether to keep the input entities:

        0 - Delete lines NL1 and NL2 and their common keypoint.
            Keypoints will not be deleted if they are meshed or if
            they are attached to other lines.  Lines will not be
            deleted if they are attached to different areas.

        1 - Keep NL1, NL2, and their common keypoint.  (The common
            keypoint will not be attached to the output line.)

    Returns
    -------
    int
        Line number of the combined line.

    Examples
    --------
    Create two lines and combine them.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> k2 = mapdl.k("", 2, 0, 0)
    >>> l0 = mapdl.l(k0, k1)
    >>> l1 = mapdl.l(k0, k2)
    >>> lout = mapdl.lcomb(l0, l1)
    >>> lout
    1

    Notes
    -----
    Combines adjacent lines into one line (the output line).  This
    operation will effectively "undo" the LDIV operation.  Line
    divisions are set to zero (use LESIZE, etc. to modify).  Lines
    attached to the same area(s) can also be combined.  See also
    the LCCAT command for line concatenation capability.
    """
    command = f"LCOMB,{nl1},{nl2},{keep}"
    return parse.parse_line_no(self.run(command, **kwargs))


def ldele(self, nl1="", nl2="", ninc="", kswp="", **kwargs):
    """Deletes unmeshed lines.

    APDL Command: LDELE

    Parameters
    ----------
    nl1, nl2, ninc
        Delete lines from NL1 to NL2 (defaults to NL1) in steps of NINC
        (defaults to 1).  If NL1 = ALL, NL2 and NINC are ignored and all
        selected lines [LSEL] are deleted.  If NL1 = P, graphical picking
        is enabled and all remaining command fields are ignored (valid only
        in the GUI).  A component name may also be substituted for NL1 (NL2
        and NINC are ignored).

    kswp
        Specifies whether keypoints are also to be deleted:

        0 - Delete lines only.

        1 - Delete lines, as well as keypoints attached to lines but not attached to other
            lines.

    Notes
    -----
    A line attached to an area cannot be deleted unless the area is first
    deleted.
    """
    command = f"LDELE,{nl1},{nl2},{ninc},{kswp}"
    return self.run(command, **kwargs)


def ldiv(self, nl1="", ratio="", pdiv="", ndiv="", keep="", **kwargs) -> str:
    """Divides a single line into two or more lines.

    APDL Command: LDIV

    Parameters
    ----------
    nl1
        Number of the line to be divided.  If negative, assume P1
        (see below) is the second keypoint of the line instead of
        the first for RATIO.  If ALL, divide all selected lines
        [LSEL].  A component name may also be substituted for NL1.

    ratio
        Ratio of line length P1-PDIV to line length P1-P2.  Must
        be between 0.0 and 1.0. Input ignored if NDIV > 2.

    pdiv
        Number to be assigned to keypoint generated at division
        location (defaults to lowest available keypoint number
        [NUMSTR]).  Input ignored if NL1 = ALL or NDIV > 2.  If
        PDIV already exists and lies on line NL1, divide line at
        PDIV (RATIO must also be 0.0).  If PDIV already exists and
        does not lie on line NL1, PDIV is projected and moved to
        the nearest point on line NL1 (if possible). PDIV cannot
        be attached to another line, area, or volume.

    ndiv
        The number of new lines to be generated from old line
        (defaults to 2).

    keep
        Specifies whether to keep the input entities:

        0 - Modify old line to use new keypoints and slopes.

        1 - Do not modify old line.  New lines will overlay old
            line and have unique keypoints.

    Returns
    -------
    str
        MAPDL command output.

    Examples
    --------
    Create a single line and divide it exactly half.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> l0 = mapdl.l(k0, k1)
    >>> output = mapdl.ldiv(l0, ratio=0.5)
    >>> print(output)
    DIVIDE LINE      1       RATIO=  0.50000       NEW KEYPOINT=     0
      NUMBER OF LINES DIVIDED =      1


    Create a single line and divide it into 5 pieces.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 0, 0)
    >>> l0 = mapdl.l(k0, k1)
    >>> mapdl.ldiv(l0, ndiv=5)

    Notes
    -----
    Divides a single line NL1 (defined from keypoint P1 to
    keypoint P2) into two or more lines.  Line NL1 becomes the new
    line beginning with keypoint P1 and new lines are generated
    ending at keypoint P2.  If the line is attached to an area,
    the area will also be updated.  Line divisions are set to zero
    (use LESIZE, etc. to modify).
    """
    command = f"LDIV,{nl1},{ratio},{pdiv},{ndiv},{keep}"
    return self.run(command, **kwargs)


def ldrag(self, nk1="", nk2="", nk3="", nk4="", nk5="", nk6="", nl1="",
          nl2="", nl3="", nl4="", nl5="", nl6="", **kwargs):
    """Generates lines by sweeping a keypoint pattern along  path.

    APDL Command: LDRAG

    Parameters
    ----------
    nk1, nk2, nk3, . . . , nk6
        List of keypoints in the pattern to be dragged (6 maximum if using
        keyboard entry).  If NK1 = P, graphical picking is enabled and all
        remaining command fields are ignored (valid only in the GUI).  If
        NK1 = ALL, all selected keypoints (except those that define the
        drag path) will be swept along the path.  A component name may also
        be substituted for NK1.

    nl1, nl2, nl3, . . . , nl6
        List of lines defining the path along which the pattern is to be
        dragged (6 maximum if using keyboard entry).  Must be a continuous
        set of lines.

    Notes
    -----
    Generates lines (and their corresponding keypoints) by sweeping a given
    keypoint pattern along a characteristic drag path.  If the drag path
    consists of multiple lines, the drag direction is determined by the
    sequence in which the path lines are input (NL1, NL2, etc.).  If the
    drag path is a single line (NL1), the drag direction is from the
    keypoint on the drag line that is closest to the first keypoint of the
    given pattern to the other end of the drag line.

    The magnitude of the vector between the keypoints of the given pattern
    and the first path keypoint remains constant for all generated keypoint
    patterns and the path keypoints.  The direction of the vector relative
    to the path slope also remains constant so that patterns may be swept
    around curves.  Keypoint and line numbers are automatically assigned
    (beginning with the lowest available values [NUMSTR]).  For best
    results, the entities to be dragged should be orthogonal to the start
    of the drag path.  Drag operations that produce an error message may
    create some of the desired entities prior to terminating.
    """
    command = f"LDRAG,{nk1},{nk2},{nk3},{nk4},{nk5},{nk6},{nl1},{nl2},{nl3},{nl4},{nl5},{nl6}"
    return self.run(command, **kwargs)


def lextnd(self, nl1="", nk1="", dist="", keep="", **kwargs) -> int:
    """Extends a line at one end by using its slope.

    APDL Command: LEXTND

    Parameters
    ----------
    nl1
        Number of the line to be extended.

    nk1
        Number of keypoint at the end of line NL1 to be extended.

    dist
        Distance that the line will be extended.

    keep
        Specifies whether to keep the input entities:

        0 - Modify old line to use new keypoints and slopes.

        1 - Do not modify old line.  New line will overlay old
            line and have unique keypoints.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create a circular arc and extend it at one of its keypoints

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 0, 1)
    >>> carcs = mapdl.circle(k0, 1, k1, arc=90)
    >>> lnum = mapdl.lextnd(carcs[0], 3, 1)
    >>> lnum
    1

    Notes
    -----
    Extends a line at one end by using its slope.  Lines may be
    redefined only if not yet attached to an area.  Line divisions
    are set to zero (use LESIZE, etc. to modify).  Note that solid
    modeling in a toroidal coordinate system is not recommended.
    """
    command = f"LEXTND,{nl1},{nk1},{dist},{keep}"
    return parse.parse_line_no(self.run(command, **kwargs))


def lfillt(self, nl1="", nl2="", rad="", pcent="", **kwargs) -> int:
    """Generate a fillet line between two intersecting lines.

    APDL Command: LFILLT

    Parameters
    ----------
    nl1
        Number of the first intersecting line.

    nl2
        Number of the second intersecting line.

    rad
        Radius of fillet to be generated.  Radius should be less than the
        lengths of the two lines specified with NL1 and NL2.

    pcent
        Number to be assigned to generated keypoint at fillet arc center.
        If zero (or blank), no keypoint is generated.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create two intersecting lines at a right angle and add a
    fillet between them.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 1, 0)
    >>> k2 = mapdl.k("", 1, 0, 0)
    >>> l0 = mapdl.l(k0, k1)
    >>> l1 = mapdl.l(k0, k2)
    >>> lnum = mapdl.lfillt(l0, l1, 0.25)
    3

    Notes
    -----
    Generates a fillet line between two intersecting lines NL1
    (P1-PINT) and NL2 (P2-PINT).  Three keypoints may be
    generated, two at the fillet tangent points (PTAN1 and PTAN2)
    and one (optional) at the fillet arc center (PCENT).  Line
    P1-PINT becomes P1-PTAN1, P2-PINT becomes P2-PTAN2, and new
    arc line PTAN1-PTAN2 is generated.  Generated keypoint and
    line numbers are automatically assigned (beginning with the
    lowest available values [NUMSTR]).  Line divisions are set to
    zero (use LESIZE, etc. to modify).
    """
    command = f"LFILLT,{nl1},{nl2},{rad},{pcent}"
    return parse.parse_line_no(self.run(command, **kwargs))


def lgen(self, itime="", nl1="", nl2="", ninc="", dx="", dy="", dz="",
         kinc="", noelem="", imove="", **kwargs):
    """Generates additional lines from a pattern of lines.

    APDL Command: LGEN

    Parameters
    ----------
    itime
        Do this generation operation a total of ITIMEs, incrementing all
        keypoints in the given pattern automatically (or by KINC) each time
        after the first.  ITIME must be > 1 for generation to occur.

    nl1, nl2, ninc
        Generate lines from pattern beginning with NL1 to NL2 (defaults to
        NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
        are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
        graphical picking is enabled and all remaining command fields are
        ignored (valid only in the GUI).  A component name may also be
        substituted for NL1 (NL2 and NINC are ignored).

    dx, dy, dz
        Keypoint location increments in the active coordinate system (--,
        Dθ, DZ for cylindrical, --, Dθ, -- for spherical).

    kinc
        Keypoint increment between generated sets.  If zero, the lowest
        available keypoint numbers are assigned [NUMSTR].

    noelem
        Specifies if elements and nodes are also to be generated:

        0 - Generate nodes and elements associated with the original lines, if they exist.

        1 - Do not generate nodes and elements.

    imove
        Specifies whether to redefine the existing lines:

        0 - Generate additional lines as requested with the ITIME argument.

        1 - Move original lines to new position retaining the same keypoint numbers (ITIME,
            KINC, and NOELM are ignored).  Valid only if the old lines are
            no longer needed at their original positions.  Corresponding
            meshed items are also moved if not needed at their original
            position.

    Notes
    -----
    Generates additional lines (and their corresponding keypoints and mesh)
    from a given line pattern.  The MAT, TYPE, REAL, and ESYS attributes
    are based upon the lines in the pattern and not upon the current
    settings.  End slopes of the generated lines remain the same (in the
    active coordinate system) as those of the given pattern.  For example,
    radial slopes remain radial, etc.  Generations which produce lines of a
    size or shape different from the pattern (i.e., radial generations in
    cylindrical systems, radial and phi generations in spherical systems,
    and theta generations in elliptical systems) are not allowed.  Note
    that solid modeling in a toroidal coordinate system is not recommended.
    New line numbers are automatically assigned (beginning with the lowest
    available values [NUMSTR]).
    """
    command = f"LGEN,{itime},{nl1},{nl2},{ninc},{dx},{dy},{dz},{kinc},{noelem},{imove}"
    return self.run(command, **kwargs)


def llist(self, nl1="", nl2="", ninc="", lab="", **kwargs):
    """Lists the defined lines.

    APDL Command: LLIST

    Parameters
    ----------
    nl1, nl2, ninc
        List lines from NL1 to NL2 (defaults to NL1) in steps of NINC
        (defaults to 1).  If NL1 = ALL (default), NL2 and NINC are ignored
        and all selected lines [LSEL] are listed.  If NL1 = P, graphical
        picking is enabled and all remaining command fields are ignored
        (valid only in the GUI).  A component name may also be substituted
        for NL1 (NL2 and NINC are ignored).

    lab
        Determines what type of listing is used (one of the following):

        (blank) - Prints information about all lines in the specified range.

        RADIUS - Prints the radius of certain circular arcs, along with the keypoint numbers of
                 each line.  Straight lines, non-circular curves, and
                 circular arcs not internally identified as arcs (which
                 depends upon how each arc is created) will print a radius
                 value of zero.

        LAYER - Prints layer-mesh control specifications.

        HPT - Prints information about only those lines that contain hard points. HPT is not
              supported in the GUI.

        ORIENT - Prints a list of lines, and identifies any orientation keypoints and any cross
                 section IDs that are associated with the lines.  Used for
                 beam meshing with defined orientation nodes and cross
                 sections.

    Notes
    -----
    There are 2 listings for the number of element divisions and the
    spacing ratio. The first listing shows assignments from LESIZE only,
    followed by the "hard" key (KYNDIV). See LESIZE for more information.
    The second listing shows NDIV and SPACE for the existing mesh, if one
    exists. Whether this existing mesh and the mesh generated by LESIZE
    match at any given point depends upon meshing options and the sequence
    of meshing operations.

    A "-1" in the "nodes" column indicates that the line has been meshed
    but that there are no interior nodes.

    An attribute (TYPE, MAT, REAL, or ESYS) listed as a zero is unassigned;
    one listed as a positive value indicates that the attribute was
    assigned with the LATT command (and will not be reset to zero if the
    mesh is cleared); one listed as a negative value indicates that the
    attribute was assigned using the attribute pointer [TYPE, MAT, REAL, or
    ESYS] that was active during meshing (and will be reset to zero if the
    mesh is cleared).

    This command is valid in any processor.
    """
    command = f"LLIST,{nl1},{nl2},{ninc},{lab}"
    return self.run(command, **kwargs)


def lplot(self, nl1="", nl2="", ninc="", **kwargs):
    """Displays the selected lines.

    APDL Command: LPLOT

    Parameters
    ----------
    nl1, nl2, ninc
        Display lines from NL1 to NL2 (defaults to NL1) in steps of NINC
        (defaults to 1).  If NL1 = ALL (default), NL2 and NINC are ignored
        and display all selected lines [LSEL].

    Notes
    -----
    Mesh divisions on plotted lines are controlled by the LDIV option of
    the /PSYMB command.

    This command is valid in any processor.
    """
    command = f"LPLOT,{nl1},{nl2},{ninc}"
    return self.run(command, **kwargs)


def lreverse(self, lnum="", noeflip="", **kwargs):
    """Reverses the normal of a line, regardless of its connectivity or mesh

    APDL Command: LREVERSE
    status.

    Parameters
    ----------
    lnum
        Line number of the line whose normal direction is to be reversed.
        If LNUM = ALL, the normals of all selected lines will be reversed.
        If LNUM = P, graphical picking is enabled.  A component name may
        also be substituted for LNUM.

    noeflip
        Indicates whether you want to change the normal direction of the
        existing elements on the reversed line(s) so that they are
        consistent with each line's new normal direction.

        0 - Make the normal direction of existing elements on the reversed line(s)
            consistent with each line's new normal direction (default).

        1 - Do not change the normal direction of existing elements on the reversed
            line(s).

    Notes
    -----
    You cannot use the LREVERSE command to change the normal direction of
    any element that has a body or surface load.  We recommend that you
    apply all of your loads only after ensuring that the element normal
    directions are acceptable.

    Real constants (such as nonuniform shell thickness and tapered beam
    constants) may be invalidated by an element reversal.

    For more information, see Revising Your Model in the Modeling and
    Meshing Guide.
    """
    command = f"LREVERSE,{lnum},{noeflip}"
    return self.run(command, **kwargs)


def lrotat(self, nk1="", nk2="", nk3="", nk4="", nk5="", nk6="", pax1="",
           pax2="", arc="", nseg="", **kwargs):
    """Generates circular lines by rotating a keypoint pattern about an axis.

    APDL Command: LROTAT

    Parameters
    ----------
    nk1, nk2, nk3, . . . , nk6
        List of keypoints in the pattern to be rotated (6 maximum if using
        keyboard entry).  If NK1 = P, graphical picking is enabled and all
        remaining command fields are ignored (valid only in the GUI).  If
        NK1 = ALL, all selected keypoints (except PAX1 and PAX2) will
        define the pattern to be rotated.  A component name may also be
        substituted for NK1.

    pax1, pax2
        Keypoints defining the axis about which the keypoint pattern is to
        be rotated.

    arc
        Arc length (in degrees).  Positive follows right-hand rule about
        PAX1-PAX2 vector.  Defaults to 360.

    nseg
        Number of lines (8 maximum) around circumference.  Defaults to
        minimum required for 90° (maximum) arcs, i.e., 4 for 360°, 3 for
        270°, etc.

    Notes
    -----
    Generates circular lines (and their corresponding keypoints) by
    rotating a keypoint pattern about an axis.  Keypoint patterns are
    generated at regular angular locations (based on a maximum spacing of
    90°).  Line patterns are generated at the keypoint patterns.  Keypoint
    and line numbers are automatically assigned (beginning with the lowest
    available values [NUMSTR]).
    """
    command = f"LROTAT,{nk1},{nk2},{nk3},{nk4},{nk5},{nk6},{pax1},{pax2},{arc},{nseg}"
    return self.run(command, **kwargs)


def lsscale(self, nl1="", nl2="", ninc="", rx="", ry="", rz="", kinc="",
            noelem="", imove="", **kwargs):
    """Generates a scaled set of lines from a pattern of lines.

    APDL Command: LSSCALE

    Parameters
    ----------
    nl1, nl2, ninc
        Set of lines (NL1 to NL2 in steps of NINC) that defines the pattern
        to be scaled.  NL2 defaults to NL1, NINC defaults to 1.  If NL1 =
        ALL, NL2 and NINC are ignored and the pattern is defined by all
        selected lines.  If NL1 = P, graphical picking is enabled and all
        remaining command fields are ignored (valid only in the GUI).  A
        component name may also be substituted for NL1 (NL2 and NINC are
        ignored).

    rx, ry, rz
        Scale factors to be applied to the X, Y, Z keypoint coordinates in
        active coordinate system (RR, Rθ, RZ for cylindrical; RR, Rθ, RΦ
        for spherical).  Note that the Rθ and RΦ scale factors are
        interpreted as angular offsets.  For example, for CSYS,1, RR, Rθ,
        RZ input of (1.5,10,3) would scale the specified keypoints 1.5
        times in the radial and 3 times in the Z direction, while adding an
        offset of 10 degrees to the keypoints.  Zero, blank, or negative
        scale factor values are assumed to be 1.0.  Zero or blank angular
        offsets have no effect.

    kinc
        Increment to be applied to keypoint numbers for generated set.  If
        zero, the lowest available keypoint numbers will be assigned
        [NUMSTR].

    noelem
        Specifies whether nodes and elements are also to be generated:

        0 - Nodes and line elements associated with the original lines will be generated
            (scaled) if they exist.

        1 - Nodes and line elements will not be generated.

    imove
        Specifies whether lines will be moved or newly defined:

        0 - Additional lines will be generated.

        1 - Original lines will be moved to new position (KINC and NOELEM are ignored).
            Use only if the old lines are no longer needed at their
            original positions.  Corresponding meshed items are also moved
            if not needed at their original position.

    Notes
    -----
    Generates a scaled set of lines (and their corresponding keypoints and
    mesh) from a pattern of lines.  The MAT, TYPE, REAL, and ESYS
    attributes are based on the lines in the pattern and not the current
    settings.  Scaling is done in the active coordinate system.  Lines in
    the pattern could have been generated in any coordinate system.
    """
    command = f"LSSCALE,{nl1},{nl2},{ninc},{rx},{ry},{rz},{kinc},{noelem},{imove}"
    return self.run(command, **kwargs)


def lstr(self, p1="", p2="", **kwargs) -> int:
    """Define a straight line irrespective of the active coordinate system.

    APDL Command: LSTR

    Parameters
    ----------
    p1
        Keypoint at the beginning of line.

    p2
        Keypoint at the end of line.

    Returns
    -------
    int
        Line number of the generated line.

    Examples
    --------
    Create a cartesian straight line regardless of the coordinate
    system being in cylindrical.

    >>> mapdl.csys(1)
    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 1, 1, 1)
    >>> lnum = mapdl.lstr(k0, k1)
    >>> lnum
    1

    Notes
    -----
    Defines a straight line from P1 to P2 using the global
    Cartesian coordinate system.  The active coordinate system
    will be ignored.  The line shape is invariant with the
    coordinate system after it is generated.  Lines may be
    redefined only if not yet attached to an area.
    """
    command = f"LSTR,{p1},{p2}"
    return parse.parse_line_no(self.run(command, **kwargs))


def lsum(self, **kwargs):
    """Calculates and prints geometry statistics of the selected lines.

    APDL Command: LSUM

    Notes
    -----
    Calculates and prints geometry statistics (length, centroid, moments of
    inertia, etc.) associated with the selected lines.  Geometry items are
    reported in the global Cartesian coordinate system.  A unit density is
    assumed, irrespective of any material associations [LATT, MAT].  Items
    calculated by LSUM and later retrieved by a *GET or *VGET command are
    valid only if the model is not modified after the LSUM command is
    issued.
    """
    command = f"LSUM,"
    return self.run(command, **kwargs)


def lsymm(self, ncomp="", nl1="", nl2="", ninc="", kinc="", noelem="",
          imove="", **kwargs):
    """Generates lines from a line pattern by symmetry reflection.

    APDL Command: LSYMM

    Parameters
    ----------
    ncomp
        Symmetry key:

        X - X symmetry (default).

        Y - Y symmetry.

        Z - Z symmetry.

    nl1, nl2, ninc
        Reflect lines from pattern beginning with NL1 to NL2 (defaults to
        NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
        are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
        graphical picking is enabled and all remaining command fields are
        ignored (valid only in the GUI).  A component name may also be
        substituted for NL1 (NL2 and NINC are ignored).

    kinc
        Keypoint increment between sets.  If zero, the lowest available
        keypoint  numbers are assigned [NUMSTR].

    noelem
        Specifies whether nodes and elements are also to be generated:

        0 - Generate nodes and elements associated with the original lines, if they exist.

        1 - Do not generate nodes and elements.

    imove
        Specifies whether areas will be moved or newly defined:

        0 - Generate additional lines.

        1 - Move original lines to new position retaining the same keypoint numbers (KINC
            and NOELEM are ignored).  Valid only if the old lines are no
            longer needed at their original positions.  Corresponding
            meshed items are also moved if not needed at their original
            position.

    Notes
    -----
    Generates a reflected set of lines (and their corresponding keypoints
    and mesh) from a given line pattern by a symmetry reflection (see
    analogous node symmetry command, NSYM).  The MAT, TYPE, REAL, and ESYS
    attributes are based upon the lines in the pattern and not upon the
    current settings.  Reflection is done in the active coordinate system
    by changing a particular coordinate sign.  The active coordinate system
    must be Cartesian.  Lines in the pattern may have been generated in any
    coordinate system.  However, solid modeling in a toroidal coordinate
    system is not recommended.  Lines are generated as described in the
    LGEN command.

    See the ESYM command for additional information about symmetry
    elements.
    """
    command = f"LSYMM,{ncomp},{nl1},{nl2},{ninc},{kinc},{noelem},{imove}"
    return self.run(command, **kwargs)


def ltan(self, nl1="", p3="", xv3="", yv3="", zv3="", **kwargs) -> int:
    """Generate a line at the end of, and tangent to, an existing line.

    APDL Command: LTAN

    Parameters
    ----------
    nl1
        Number of the line the generated line is tangent to.  If
        negative, assume P1 (see below), instead of P2, is the
        second keypoint of line NL1.

    p3
        Keypoint at which generated line must end.

    Returns
    -------
    int
        Line number of the line generated.

    Examples
    --------
    Create a circular arc and generate a tangent spline at the end
    of it directed to a new keypoint.

    >>> k0 = mapdl.k("", 0, 0, 0)
    >>> k1 = mapdl.k("", 0, 0, 1)
    >>> k2 = mapdl.k("", -1, 1.5, 0)
    >>> carc = mapdl.circle(k0, 1, k1, arc=90)
    >>> lnum = mapdl.ltan(carc[0], k2)
    >>> lnum
    2

    Notes
    -----
    Generates a line (P2-P3) tangent at end point (P2) of line NL1
    (P1-P2).
    """
    command = f"LTAN,{nl1},{p3},{xv3},{yv3},{zv3}"
    return parse.parse_line_no(self.run(command, **kwargs))


def ltran(self, kcnto="", nl1="", nl2="", ninc="", kinc="", noelem="",
          imove="", **kwargs):
    """Transfers a pattern of lines to another coordinate system.

    APDL Command: LTRAN

    Parameters
    ----------
    kcnto
        Reference number of coordinate system where the pattern is to be
        transferred.  Transfer occurs from the active coordinate system.
        The coordinate system type and parameters of KCNTO must be the same
        as the active system.

    nl1, nl2, ninc
        Transfer lines from pattern beginning with NL1 to NL2 (defaults to
        NL1) in steps of NINC (defaults to 1).  If NL1 = ALL, NL2 and NINC
        are ignored and pattern is all selected lines [LSEL].  If NL1 = P,
        graphical picking is enabled and all remaining command fields are
        ignored (valid only in the GUI).  A component name may also be
        substituted for NL1 (NL2 and NINC are ignored).

    kinc
        Keypoint increment between sets.  If zero, the lowest available
        keypoint numbers are assigned [NUMSTR].

    noelem
        Specifies whether nodes and elements are also to be generated:

        0 - Generate nodes and elements associated with the original lines, if they exist.

        1 - Do not generate nodes and elements.

    imove
        Specifies whether lines will be moved or newly defined:

        0 - Generate additional lines.

        1 - Move original lines to new position retaining the same keypoint numbers (KINC
            and NOELM are ignored).  Valid only if the old lines are no
            longer needed at their original positions.  Corresponding
            meshed items are also moved if not needed at their original
            position.

    Notes
    -----
    Transfers a pattern of lines (and their corresponding keypoints and
    mesh) from one coordinate system to another (see analogous node
    transfer command, TRANSFER).  The MAT, TYPE, REAL, and ESYS attributes
    are based upon the lines in the pattern and not upon the current
    settings.  Coordinate systems may be translated and rotated relative to
    each other.  Initial pattern may be generated in any coordinate system.
    However, solid modeling in a toroidal coordinate system is not
    recommended.  Coordinate and slope values are interpreted in the active
    coordinate system and are transferred directly.  Lines are generated as
    described in the LGEN command.
    """
    command = f"LTRAN,{kcnto},{nl1},{nl2},{ninc},{kinc},{noelem},{imove}"
    return self.run(command, **kwargs)


def spline(self, p1="", p2="", p3="", p4="", p5="", p6="", xv1="", yv1="",
           zv1="", xv6="", yv6="", zv6="", **kwargs) -> list:
    """Generate a segmented spline through a series of keypoints.

    APDL Command: SPLINE

    Parameters
    ----------
    p1, p2, p3, . . . , p6
        Keypoints through which the spline is fit.  At least two
        must be defined.

    Returns
    -------
    list
        List of line numbers generated.

    Examples
    --------
    Create a spline with 5 keypoints.

    >>> k0 = mapdl.k('', 0, 0, 0)
    >>> k1 = mapdl.k('', 0.2, 0.2, 0)
    >>> k2 = mapdl.k('', 0.4, 0.3, 0)
    >>> k3 = mapdl.k('', 0.6, 0.5, 0)
    >>> k4 = mapdl.k('', 0.8, 0.3, 0)
    >>> lines = mapdl.spline(k0, k1, k2, k3, k4)
    >>> lines
    [1, 2, 3, 4]

    Notes
    -----
    The output from this command is a series of connected lines
    (one line between each pair of keypoints) that together form a
    spline.  Note that solid modeling in a toroidal coordinate
    system is not recommended.
    """
    command = f"SPLINE,{p1},{p2},{p3},{p4},{p5},{p6},{xv1},{yv1},{zv1},{xv6},{yv6},{zv6}"
    return parse.parse_line_nos(self.run(command, **kwargs))


def ssln(self, fact="", size="", **kwargs):
    """Selects and displays small lines in the model.

    APDL Command: SSLN

    Parameters
    ----------
    fact
        Factor used to determine small lines.  FACT times the average line
        length in the model is used as the line length limit below which
        lines will be selected.

    size
        Line length limit for line selection.  Lines that have a length
        less than or equal to SIZE will be selected.  Used only if FACT is
        blank.

    Notes
    -----
    SSLN invokes a predefined ANSYS macro for selecting small lines in a
    model.  Lines that are smaller than or equal to the specified limit
    (FACT or SIZE) are selected and line numbers are displayed.  This
    command macro is useful for detecting very small lines in a model that
    may cause problems (i.e., poorly shaped elements or a meshing failure)
    during meshing.  All lines that are not "small" will be unselected and
    can be reselected with the LSEL command.
    """
    command = f"SSLN,{fact},{size}"
    return self.run(command, **kwargs)



def axpy(self, vr="", vi="", m1="", wr="", wi="", m2="", **kwargs):
    """Performs the matrix operation M2= v*M1 + w*M2.

    APDL Command: *AXPY

    Parameters
    ----------
    vr, vi
        The real and imaginary parts of the scalar v. Default value is 0.

    m1
        Name of matrix M1. If not specified, the operation M2 = w*M2 will
        be performed.

    wr, wi
        The real and imaginary parts of the scalar w. Default value is 0.

    m2
        Name of matrix M2. Must be specified.

    Notes
    -----
    The matrices M1 and M2 must have the same dimensions and same type
    (dense or sparse). If M2 is real, vi and wi are ignored.
    """
    command = f"*AXPY,{vr},{vi},{m1},{wr},{wi},{m2}"
    return self.run(command, **kwargs)


def comp(self, matrix="", algorithm="", threshold="", **kwargs):
    """Compresses the columns of a matrix using a specified algorithm.

    APDL Command: *COMP

    Parameters
    ----------
    matrix
        Name of the matrix to compress.

    algorithm
        Algorithm to use:

        Singular value decomposition algorithm (default). - Modified Gram-Schmidt algorithm.

    threshold
        Numerical threshold value used to manage the compression. Default
        value for SVD is 1E-7; default value for MGS is 1E-14.

    Notes
    -----
    The algorithms available through this command are only applicable to
    dense matrices that were created using the *DMAT command.

    Columns which are linearly dependent on others are removed, leaving the
    independent or basis vectors. The matrix is resized according to the
    new size determined by the algorithm.
    """
    command = f"*COMP,{matrix},{algorithm},{threshold}"
    return self.run(command, **kwargs)


def dmat(self, matrix="", type_="", method="", val1="", val2="", val3="",
         val4="", val5="", **kwargs):
    """Creates a dense matrix.

    APDL Command: *DMAT

    Parameters
    ----------
    matrix
        Name used to identify the matrix. Must be specified.

    type\_
        Matrix type:

        Double precision real values (default). - Complex double precision values.

    method
        Method used to create the matrix:

        Allocate space for a matrix (default). - Resize an existing matrix to new row and column dimensions. Values are kept
                          from the original matrix. If the dimensions
                          specified by Val1 (rows) and Val2 (columns) are
                          greater than the original matrix size, the
                          additional entries are assigned a value of zero.

        Copy an existing matrix. - Link to an existing matrix. The memory will be shared between the original
                          matrix and the new matrix. This is useful for
                          manipulating a submatrix of a larger matrix. The
                          Val1 through Val5 arguments will be used to
                          specify the lower and upper bounds of row and
                          column numbers from the original matrix.

    val1, val2, val3, val4, val5
        Additional input. The meaning of Val1 through Val5 will vary
        depending on the specified Method. See details below.

    Notes
    -----
    This command allows you to create a dense matrix. To create a sparse
    matrix, use the *SMAT command. *SMAT is recommended for large matrices
    obtained from the .FULL or .HBMAT file. Refer to the HBMAT command
    documentation for more information about .FULL file contents.

    Use the *VEC command to create a vector.

    For very large matrices, use the OUTOFCORE option (Method = ALLOC or
    COPY) to keep some of the matrix on disk if there is insufficient
    memory.

    When importing a dense matrix from a DMIG file, you can define the
    formatting of the file using the Val3 and Val4 fields. Here are a few
    different example of formats:

    A formatted file (using Val3=’F’, and Val4=8):
    """
    command = f"*DMAT,{matrix},{type_},{method},{val1},{val2},{val3},{val4},{val5}"
    return self.run(command, **kwargs)


def dot(self, vector1="", vector2="", par_real="", par_imag="", **kwargs):
    """Computes the dot (or inner) product of two vectors.

    APDL Command: *DOT

    Parameters
    ----------
    vector1
        Name of first vector; must have been previously specified by a *VEC
        command.

    vector2
        Name of second vector; must have been previously specified by a
        *VEC command.

    par_real
        Parameter name that contains the result.

    par_imag
        Parameter name that contains the imaginary part of the result (used
        only for complex vectors).

    Notes
    -----
    If Vector1 and Vector2 are complex, the complex conjugate of Vector1 is
    used to compute the result (Par_Real, Par_Imag).
    """
    command = f"*DOT,{vector1},{vector2},{par_real},{par_imag}"
    return self.run(command, **kwargs)


def eigen(self, kmatrix="", mmatrix="", cmatrix="", evals="", evects="",
          **kwargs):
    """Performs a modal solution with unsymmetric or damping matrices.

    APDL Command: *EIGEN

    Parameters
    ----------
    kmatrix
        Name of the stiffness matrix. May be a real or complex-valued
        matrix.

    mmatrix
        Name of the mass matrix.

    cmatrix
        Name of the damping matrix (used only for MODOPT,DAMP).

    evals
        Name of the output eigenvalues vector. It will be an m-long *VEC
        vector of complex values, where m is the number of eigenvalues
        requested (MODOPT).

    evects
        Name of the output eigenvector matrix. It will be a n x m *DMAT
        (dense) matrix of complex values, where n is the size of the matrix
        and m is the number of eigenvalues requested (MODOPT).

    Notes
    -----
    Use the command ANTYPE,MODAL and the MODOPT command to specify the
    modal solution options. Only MODOPT,DAMP, MODOPT,UNSYM, MODOPT,LANB,
    and MODOPT,SUBSP are supported.

    *EIGEN with Block Lanczos (LANB) only supports sparse matrices.

    Distributed ANSYS Restriction: This command is not supported in
    Distributed ANSYS.
    """
    command = f"*EIGEN,{kmatrix},{mmatrix},{cmatrix},{evals},{evects}"
    return self.run(command, **kwargs)


def export(self, matrix="", format_="", fname="", val1="", val2="", val3="",
           **kwargs):
    """Exports a matrix to a file in the specified format.

    APDL Command: *EXPORT

    Parameters
    ----------
    matrix
        Name of the matrix to export (must be a matrix previously created
        with *DMAT or *SMAT, or a vector previously created with *VEC).

    format\_
        Format of the output file:

        Export the matrix in the Matrix Market Format. - Export the matrix in the SUB file format.

        Export the matrix in the Harwell-Boeing file format. - Export the matrix in a native format, to be re-imported using the *DMAT or
                          *SMAT command.

        Export the matrix to an existing EMAT file. - Export the matrix to an APDL array parameter.

        Export the matrix profile to a Postscript file. - Export the matrix in the DMIG file format.

    fname
        Name of the file, or name of the array parameter if Format = APDL.

    val1, val2, val3
        Additional input. The meaning of Val1 through Val3 will vary
        depending on the specified Format. See table below for details.

    Notes
    -----
    Only sparse matrices can be exported to Postscript files. This option
    plots the matrix profile as a series of dots.

    If you want to create a .SUB file from several matrices, you need to
    set Val3 = WAIT for all matrices but the last, and Val3 = DONE for the
    last one. The export will be effective at the last *EXPORT command.

    To create a .SUB file or .DMIG file from scratch, you must supply the
    row information array. (Specify this array in the Val2 field for .SUB
    or in the Val1 field for .DMIG.) This must be an m x 2 array, where m
    is the size of the matrix. The first column is the node number and the
    second column is the DOF number corresponding to each row of the
    matrix.

    The *EXPORT command is not applicable to sparse matrices initialized
    from .FULL files by means of the NOD2BCS option on the *SMAT command
    (i.e., *SMAT,,,IMPORT,FULL,,NOD2BCS).
    """
    command = f"*EXPORT,{matrix},{format_},{fname},{val1},{val2},{val3}"
    return self.run(command, **kwargs)


def fft(self, type_="", inputdata="", outputdata="", dim1="", dim2="",
        resultformat="", **kwargs):
    """Computes the fast Fourier transformation of a specified matrix or

    APDL Command: *FFT
    vector.

    Parameters
    ----------
    type\_
        Type of FFT transformation:

        Forward FFT computation (default). - Backward FFT computation.

    inputdata
        Name of matrix or vector for which the FFT will be computed. This
        can be a dense matrix (created by the *DMAT command) or a vector
        (created by the *VEC command). Data can be real or complex values.
        There is no default value for this argument.

    outputdata
        Name of matrix or vector where the FFT results will be stored. The
        type of this argument must be consistent with InputData (see table
        below). There is no default value for this argument.

    dim1
        The number of terms to consider for a vector, or the number of rows
        for a matrix. Defaults to the whole input vector or all the rows of
        the matrix.

    dim2
        The number of columns to consider for a matrix. Defaults to all the
        columns of the matrix. (Valid only for matrices.)

    resultformat
        Specifies the result format:

        Returns the full result. That is, the result matches the dimension specified on this command (DIM1, DIM2). - Returns partial results. For real input data, there is a symmetry in the
                          results of the Fourier transform as some
                          coefficients are conjugated. The partial format
                          uses this symmetry to optimize the storage of the
                          results. (Valid only for real data.)

    Notes
    -----
    In the example that follows, the fast Fourier transformation is used to
    filter frequencies from a noisy input signal.
    """
    command = f"*FFT,{type_},{inputdata},{outputdata},{dim1},{dim2},{resultformat}"
    return self.run(command, **kwargs)


def free(self, name="", **kwargs):
    """Deletes a matrix or a solver object and frees its memory allocation.

    APDL Command: *FREE

    Parameters
    ----------
    name
        Name of the matrix or solver object to delete. Use Name = ALL to
        delete all APDL Math matrices and solver objects.  Use Name = WRK
        to delete all APDL Math matrices and solver objects that belong to
        a given workspace.

    val1
        If Name = WRK, Val1 is to set the memory workspace number.

    Notes
    -----
    A /CLEAR command will automatically delete all the current APDL Math
    objects.
    """
    command = f"*FREE,{name}"
    return self.run(command, **kwargs)


def init(self, name="", method="", val1="", val2="", val3="", **kwargs):
    """Initializes a vector or dense matrix.

    APDL Command: *INIT

    Parameters
    ----------
    name
        Vector or matrix which will be initialized. This can be a vector
        (created by the *VEC command) or a dense matrix (created by the
        *DMAT command).

    method
        Initialization method to use:

        Fill the vector/matrix with zeros (default). - Fill the vector/matrix with a constant value.

        Fill the vector/matrix with random values. - Fill the nth diagonal of the matrix with a constant value. Other values are not
                          overwritten.

    val1, val2, val3
        Additional input. The meaning of Val1 through Val3 will vary
        depending on the specified Method. See details below.

    Notes
    -----
    This command initializes a previously defined vector (*VEC) or dense
    matrix (*DMAT).
    """
    command = f"*INIT,{name},{method},{val1},{val2},{val3}"
    return self.run(command, **kwargs)


def itengine(self, type_="", enginename="", precondname="", matrix="",
             rhsvector="", solvector="", maxiter="", toler="", **kwargs):
    """Performs a solution using an iterative solver.

    APDL Command: *ITENGINE

    Parameters
    ----------
    type\_
        Specifies the algorithm to be used:

    enginename
        Name used to identify this iterative solver engine. Must be
        specified.

    precondname
        Linear solver engine name (*LSENGINE) identifying the factored
        matrix to be used as the preconditioner.

    matrix
        Name of the matrix to solve.

    rhsvector
        Matrix (load vector) name.

    solvector
        Solution vector name. If non-zero, it will be taken as the initial
        vector for the iterative process.

    maxiter
        Maximum number of iterations allowed. Default is 2 times the number
        of rows in the matrix.

    toler
        Convergence tolerance. Default is 1.0E-8.

    Notes
    -----
    This command solves Ax = b using a preconditioned conjugate gradient
    algorithm. It uses an existing factored system as the preconditioner.
    This solution method is useful if an existing matrix has been solved
    and minor changes have been made to the matrix.
    """
    command = f"*ITENGINE,{type_},{enginename},{precondname},{matrix},{rhsvector},{solvector},{maxiter},{toler}"
    return self.run(command, **kwargs)


def lsbac(self, enginename="", rhsvector="", solvector="", **kwargs):
    """Performs the solve (forward/backward substitution) of a factorized

    APDL Command: *LSBAC
    linear system.

    Parameters
    ----------
    enginename
        Name used to identify this engine. Must have been previously
        created using *LSENGINE and factorized using *LSFACTOR.

    rhsvector
        Name of vector containing the right-hand side (load) vectors as
        input. Must have been previously defined as a *VEC vector or a
        *DMAT matrix.

    solvector
        Name of vector that will contain the solution vectors upon
        completion. Must be predefined as a *VEC vector or *DMAT matrix.

    Notes
    -----
    This command performs forward and back substitution to obtain the
    solution to the linear matrix equation Ax = b. The matrix engine must
    have been previously defined using *LSENGINE, and the matrix factored
    using *LSFACTOR.

    You can use the *DMAT,,,COPY (or *VEC,,,COPY) command to copy the load
    vector to the solution vector in order to predefine it with the
    appropriate size.
    """
    command = f"*LSBAC,{enginename},{rhsvector},{solvector}"
    return self.run(command, **kwargs)


def lsdump(self, enginename="", filename="", **kwargs):
    """Dumps a linear solver engine to a binary File.

    APDL Command: *LSDUMP

    Parameters
    ----------
    enginename
        Name used to identify this engine. Must have been previously
        created using *LSENGINE and factorized using *LSFACTOR.

    filename
        Name of the file to create.

    Notes
    -----
    Dumps a previously factorized linear solver system to a binary file.
    Only LAPACK and BCS linear solvers can be used with this feature. The
    Linear Solver can later be restored with the *LSRESTORE command.

    A BCS Sparse Solver can be dumped only if uses the INCORE memory option
    (see BCSOPTION).
    """
    command = f"*LSDUMP,{enginename},{filename}"
    return self.run(command, **kwargs)


def lsengine(self, type_="", enginename="", matrix="", option="", **kwargs):
    """Creates a linear solver engine.

    APDL Command: *LSENGINE

    Parameters
    ----------
    type_
        Specifies the algorithm to be used:

        * ``"BCS"`` - Boeing sparse solver (default if applied to
                    sparse matrices).

        * ``"DSS"`` : MKL sparse linear solver (Intel Windows and
                    Linux systems only).

        * ``"LAPACK"`` : LAPACK dense matrix linear solver (default if
                       applied to dense matrices).

        * ``"DSP"`` : Distributed sparse solver.

    enginename
        Name used to identify this engine. Must be specified.

    matrix
        Name of the matrix to solve.

    option
        Option to control the memory mode of the DSS solver (used only
        if ``type='dss'``):

        * ``"INCORE"`` : In-core memory mode

        * ``"OUTOFCORE"`` : Out-of-core memory mode

    Notes
    -----
    This command creates a linear solver engine.

    The BCS, DSS, and DSP solvers can only be used with sparse matrices.
    For dense matrices, use the LAPACK solver.
    """
    command = f"*LSENGINE,{type_},{enginename},{matrix},{option}"
    return self.run(command, **kwargs)


def lsfactor(self, enginename="", option="", **kwargs):
    """Performs the numerical factorization of a linear solver system.

    APDL Command: *LSFACTOR

    Parameters
    ----------
    enginename
        Name used to identify this engine. Must have been previously
        created using *LSENGINE.

    option
        Option to invert the matrix, used only with an LAPACK engine
        (*LSENGINE,LAPACK):

    Notes
    -----
    Performs the computationally intensive, memory intensive factorization
    of a matrix specified by *LSENGINE, using the solver engine also
    specified by *LSENGINE.
    """
    command = f"*LSFACTOR,{enginename},{option}"
    return self.run(command, **kwargs)


def lsrestore(self, enginename="", filename="", **kwargs):
    """Restores a linear solver engine from a binary file.

    APDL Command: *LSRESTORE

    Parameters
    ----------
    enginename
        Name used to identify this engine.

    filename
        Name of the file to read from.

    Notes
    -----
    Restores a previously dumped Linear Solver (see the *LSDUMP command).
    This Linear Solver can be used to solve a linear system using the
    *LSBAC command.
    """
    command = "*LSRESTORE,%s,%s" % (str(enginename), str(filename))
    return self.run(command, **kwargs)


def merge(self, name1="", name2="", val1="", val2="", **kwargs):
    """Merges two dense matrices or vectors into one.

    APDL Command: *MERGE

    Parameters
    ----------
    name1
        Name of the matrix or vector to extend.

    name2
        Name of the matrix or vector to be merged into ``name1``.

    val1
        If ``name1`` refers to a dense matrix created by the *DMAT
        command then the column or row number indicating where the new values
        are to be inserted into the Name1 matrix.

        If ``name` refers to a vector created by *VEC then this is the
        row number indicating where the new values are to be inserted
        into the ``name1`` vector.

    val2
        Specifies how the ``name2`` matrix or vector is copied into
        the ``name1`` matrix.

        * ``"COL"`` : Insert the new values at the column location
          specified by ``val1`` (default).
        * ``"row"`` : Insert the new values at the row location specified by ``val1``.

    Notes
    -----
    ``merge`` can be used to add new columns or rows to a dense matrix
    that was created by the *DMAT command. In this case, ``name1`` must
    be the name of the dense matrix and ``name2`` must refer to a vector
    or another dense matrix.

    *MERGE can also be used to add new rows to a vector that was
     created by the *VEC command. In this case, ``name1`` and
     ``name2`` must both refer to vectors.

    In all cases, the values of the original matrix or vector are
    retained, and the matrix or vector is resized to accommodate the
    additional rows or columns.
    """
    return self.run(f"MERGE,{name1},{name2},{val1},{val2}", **kwargs)


def mult(self, m1="", t1="", m2="", t2="", m3="", **kwargs):
    """Performs the matrix multiplication M3 = M1(T1)*M2(T2).

    APDL Command: *MULT

    Parameters
    ----------
    m1
        Name of matrix M1. Must have been previously specified by a *DMAT
        or *SMAT command.

    t1
        Transpose key. Set T1 = TRANS to use the transpose of M1. If blank,
        transpose will not be used.

    m2
        Name of matrix M2. Must have been previously specified by a *DMAT
        command.

    t2
        Transpose key. Set T2 = TRANS to use the transpose of M2. If blank,
        transpose will not be used.

    m3
        Name of resulting matrix, M3. Must be specified.

    Notes
    -----
    The matrices must be dimensionally consistent such that the number of
    columns of M1 (or the transposed matrix, if requested) is equal to the
    number of rows of M2 (or the transposed matrix, if requested).

    You cannot multiply two sparse matrices with this command (that is, M1
    and M2 cannot both be sparse). The resulting matrix, M3, will always be
    a dense matrix, no matter what combination of input matrices is used
    (dense*sparse, sparse*dense, or dense*dense).
    """
    command = f"*MULT,{m1},{t1},{m2},{t2},{m3}"
    return self.run(command, **kwargs)


def nrm(self, name="", normtype="", parr="", normalize="", **kwargs):
    """Computes the norm of the specified matrix or vector.

    APDL Command: *NRM

    Parameters
    ----------
    name
        Matrix or vector for which the norm will be computed. This can be a
        dense matrix (created by the *DMAT command), a sparse matrix
        (created by the *SMAT command) or a vector (created by the *VEC
        command)

    normtype
        Mathematical norm to use:

        L2 (Euclidean or SRSS) norm (default). - L1 (absolute sum) norm (vectors only).

    parr
        Parameter name that contains the result.

    normalize
        Normalization key; to be used only for vectors created by *VEC:

        Normalize the vector such that the norm is 1.0. - Do not normalize the vector (default).

    Notes
    -----
    The NRM2 option corresponds to the Euclidean or L2 norm and is
    applicable to either vectors or matrices. The NRM1 option corresponds
    to the L1 norm and is applicable to vectors only. The NRMINF option is
    the maximum norm and is applicable to either vectors or matrices.
    """
    command = f"*NRM,{name},{normtype},{parr},{normalize}"
    return self.run(command, **kwargs)


def remove(self, name="", val1="", val2="", val3="", **kwargs):
    """Suppresses rows or columns of a dense matrix or a vector.

    APDL Command: *REMOVE

    Parameters
    ----------
    name
        Name of the matrix or vector to be revised.

    val1
        First row or column number to suppress if ``name`` is a dense
        matrix.  First value index to suppress if ``name`` is a
        vector.

    Val2
        Last row or column number to suppress if ``name`` is a dense
        matrix.  Last value index to suppress if ``name`` is a
        vector.

    Val3
        Specifies what to remove if ``name`` is a dense matrix.

        * ``"COL"`` : Remove columns of the matrix (default).

        * ``"ROW"`` : Remove rows of the matrix.

    Notes
    -----
    The values of the original matrix or vector specified by Name are
    retained. The matrix or vector is resized to the new number of
    rows and columns.
    """
    return self.run(f"REMOVE,{name},{val1},{val2},{val3}", **kwargs)


def scal(self, name="", val1="", val2="", **kwargs):
    """Scales a vector or matrix by a constant.

    APDL Command: *SCAL

    Parameters
    ----------
    name
        Name used to identify the vector or matrix to be scaled. Must
        be specified.

    val1
        The real part of the constant to use (default = 1).

    val2
        The imaginary part of the constant to use (default = 0). This
        value is used only if the vector or matrix specified by Name
        is complex.

    Notes
    -----
    This command can be applied to vectors and matrices created by the
    *VEC, *DMAT and *SMAT commands.
    """
    return self.run(f"*SCAL,{name},{val1},{val2}", **kwargs)


def smat(self, matrix="", type_="", method="", val1="", val2="", val3="", val4="",
         **kwargs):
    """Creates a sparse matrix.

    APDL Command: *SMAT

    Parameters
    ----------
    matrix
        Name used to identify the matrix. Must be specified.

    type\_
        Matrix type:

        Double precision real values (default). - Complex double precision values.

    method
        Method used to create the matrix:

        Copy an existing matrix. - Import the matrix from a file.

    val1, val2, val3, val4
        Additional input. The meaning of Val1 through Val3 will
        vary depending on the specified Method. See in your ansys
        documentation.

    Notes
    -----
    Use the *DMAT command to create a dense matrix.

    Unlike the *DMAT command, the *SMAT command cannot be used to allocate
    a sparse matrix.

    For more information on the NOD2BCS and USR2BCS mapping vectors, see
    Degree of Freedom Ordering in the ANSYS Parametric Design Language
    Guide.

    For more information about .FULL file contents, see the HBMAT in the
    Command Reference.
    """
    command = f"*SMAT,{matrix},{type_},{method},{val1},{val2},{val3},{val4}"
    return self.run(command, **kwargs)


def starprint(self, matrix="", fname="", **kwargs):
    """Prints the matrix values to a file.

    APDL Command: *PRINT

    Parameters
    ----------
    matrix
        Name of matrix or vector to print. Must be specified.

    fname
        File name. If blank, matrix is written to the output file.

    Notes
    -----
    The matrix may be a dense matrix (*DMAT), a sparse matrix (*SMAT), or a
    vector (*VEC). Only the non-zero entries of the matrix are printed.
    """
    command = f"*PRINT,{matrix},{fname}"
    return self.run(command, **kwargs)


def sort(self, **kwargs):
    """Specifies "Sort settings" as the subsequent status topic.

    APDL Command: SORT

    Notes
    -----
    This is a status [STAT] topic command.  Status topic commands are
    generated by the GUI and will appear in the log file (Jobname.LOG) if
    status is requested for some items under Utility Menu> List> Status.
    This command will be immediately followed by a STAT command, which will
    report the status for the specified topic.

    If entered directly into the program, the STAT command should
    immediately follow this command.
    """
    command = f"SORT,"
    return self.run(command, **kwargs)


def vec(self, vector="", type_="", method="", val1="", val2="", val3="",
        val4="", **kwargs):
    """Creates a vector.

    APDL Command: *VEC

    Parameters
    ----------
    vector
        Name used to identify the vector. Must be specified.

    type_
        Vector type:

        * ``"D"`` : Double precision real values (default).

        * ``"Z"`` : Complex double precision values.

        * ``"I"`` : Integer values.

    method
        Method used to create the vector:

        * ``"ALLOC"`` : Allocate space for a vector (default).

        * ``"RESIZE"`` : Resize an existing vector to a new
          length. Values are kept from the original vector. If the
          length specified by Val1 is greater than the original vector
          length, the additional rows are assigned a value of zero.

        * ``"COPY"`` : Copy an existing vector.

        * ``"IMPORT"`` : Import the vector from a file.

        * ``"LINK"`` : Link to a column of an existing dense *DMAT
          matrix and use it in subsequent vector calculations. Any
          changes to the vector are also made to the corresponding
          matrix column (memory is shared).

        Copy an existing vector. - Import the vector from a file.

    val1, val2, val3, val4, val5
        Additional input. The meaning of ``val1`` through ``val5`` will vary
        depending on the specified Method.  See:
        https://www.mm.bme.hu/~gyebro/files/ans_help_v182/ans_cmd/Hlp_C_VEC.html

    Notes
    -----
    Use the *DMAT command to create a matrix.

    For more information on the BACK and FORWARD nodal mapping vectors, see
    Degree of Freedom Ordering in the ANSYS Parametric Design Language
    Guide.
    """
    command = f"*VEC,{vector},{type_},{method},{val1},{val2},{val3},{val4}"
    return self.run(command, **kwargs)


def wrk(self, num="", **kwargs):
    """APDL Command: *WRK

    Sets the active workspace number.

    Parameters
    ----------
    num
        Number of the active memory workspace for APDLMath vector and
        matrices. All the following APDLMath vectors and matrices will
        belong to this memory workspace, until the next call to the *WRK
        command. By default, all the APDLMath objects belong to workspace
        number 1.

    Notes
    -----
    This feature enables you to associate a set of vector and matrices in a
    given memory workspace, so that you can easily manage the free step:

    This feature can be useful to free all the temporary APDLMath variables
    inside a MACRO in one call.
    """
    command = "*WRK,%s" % (str(num))
    return self.run(command, **kwargs)



def sfa(self, area="", lkey="", lab="", value="", value2="", **kwargs):
    """Specifies surface loads on the selected areas.

    APDL Command: SFA

    Parameters
    ----------
    area
        Area to which surface load applies.  If ALL, apply load to all
        selected areas [ASEL].  If Area = P, graphical picking is enabled
        and all remaining command fields are ignored (valid only in the
        GUI).  A component may be substituted for Area.

    lkey
        Load key associated with surface load (defaults to 1).  Load keys
        (1,2,3, etc.) are listed under "Surface Loads" in the input data
        table for each element type in the Element Reference.  LKEY is
        ignored if the area is the face of a volume region meshed with
        volume elements.

    lab
        Valid surface load label.  Load labels are listed under "Surface
        Loads" in the input table for each area type in the Element
        Reference.

    value
        Surface load value or table name reference for specifying tabular
        boundary conditions.

    value2
        Second surface load value (if any).

    Notes
    -----
    Surface loads may be transferred from areas to elements with the SFTRAN
    or SBCTRAN commands.  See the SFGRAD command for an alternate tapered
    load capability.

    Tabular boundary conditions (VALUE = %tabname% and/or VALUE2 =
    %tabname%) are available for the following surface load labels (Lab)
    only: : PRES (real and/or imaginary components), CONV (film coefficient
    and/or bulk temperature) or HFLUX, and RAD (surface emissivity and
    ambient temperature). Use the *DIM command to define a table.

    This command is also valid in PREP7.
    """
    command = f"SFA,{area},{lkey},{lab},{value},{value2}"
    return self.run(command, **kwargs)


def sfadele(self, area="", lkey="", lab="", **kwargs):
    """Deletes surface loads from areas.

    APDL Command: SFADELE

    Parameters
    ----------
    area
        Area to which surface load deletion applies.  If ALL, delete load
        from all selected areas [ASEL].  If AREA = P, graphical picking is
        enabled and all remaining command fields are ignored (valid only in
        the GUI).  A component name may be substituted for AREA.

    lkey
        Load key associated with surface load (defaults to 1).  See the SFA
        command for details.

    lab
        Valid surface load label.  If ALL, use all appropriate labels.  See
        the SFA command for labels.

    Notes
    -----
    Deletes surface loads (and all corresponding finite element loads) from
    selected areas.

    This command is also valid in PREP7.
    """
    command = f"SFADELE,{area},{lkey},{lab}"
    return self.run(command, **kwargs)


def sfalist(self, area="", lab="", **kwargs):
    """Lists the surface loads for the specified area.

    APDL Command: SFALIST

    Parameters
    ----------
    area
        Area at which surface load is to be listed.  If ALL (or blank),
        list for all selected areas [ASEL].  If AREA = P, graphical picking
        is enabled and all remaining command fields are ignored (valid only
        in the GUI).  A component name may be substituted for AREA.

    lab
        Valid surface load label.  If ALL (or blank), use all appropriate
        labels.  See the SFA command for labels.

    Notes
    -----
    This command is valid in any processor.
    """
    command = f"SFALIST,{area},{lab}"
    return self.run(command, **kwargs)


def sfl(self, line="", lab="", vali="", valj="", val2i="", val2j="",
        **kwargs):
    """Specifies surface loads on lines of an area.

    APDL Command: SFL

    Parameters
    ----------
    line
        Line to which surface load applies.  If ALL, apply load to all
        selected lines [LSEL].  If Line = P, graphical picking is enabled
        and all remaining command fields are ignored (valid only in the
        GUI).  A component name may be substituted for Line.

    lab
        Valid surface load label.  Load labels are listed under "Surface
        Loads" in the input table for each element type in the Element
        Reference.

    vali, valj
        Surface load values at the first keypoint (VALI) and at the second
        keypoint (VALJ) of the line, or table name for specifying tabular
        boundary conditions.  If VALJ is blank, it defaults to VALI.  If
        VALJ is zero, a zero is used.  If Lab = CONV, VALI and VALJ are the
        film coefficients and VAL2I and VAL2J are the bulk temperatures.
        To specify a table, enclose the table name in percent signs (%),
        e.g., %tabname%.  Use the *DIM command to define a table.  If Lab =
        CONV and VALI = -N, the film coefficient may be a function of
        temperature and is determined from the HF property table for
        material N [MP].  If Lab = RAD, VALI and VALJ values are surface
        emissivities and VAL2I and VAL2J are ambient temperatures.  The
        temperature used to evaluate the film coefficient is usually the
        average between the bulk and wall temperatures, but may be user
        defined for some elements.  If Lab = RDSF, VALI is the emissivity
        value; the following condition apply: If VALI = -N, the emissivity
        may be a function of the temperature and is determined from the
        EMISS property table for material N [MP]. If Lab = FSIN in a Multi-
        field solver (single or multiple code coupling) analysis, VALI is
        the surface interface number. If Lab = FSIN in a unidirectional
        ANSYS to CFX analysis, VALJ is the surface interface number (not
        available from within the GUI) and VALI is not used unless the
        ANSYS analysis is performed using the Multi-field solver.

    val2i, val2j
        Second surface load values (if any).  If Lab = CONV, VAL2I and
        VAL2J are the bulk temperatures. If Lab = RAD, VAL2I and VAL2J are
        the ambient temperatures. If Lab = RDSF, VAL2I is the enclosure
        number. Radiation will occur between surfaces flagged with the same
        enclosure numbers. If the enclosure is open, radiation will occur
        to the ambient. VAL2I and VAL2J are not used for other surface load
        labels.  If VAL2J is blank, it defaults to VAL2I.  If VAL2J is
        zero, a zero is used.  To specify a table (Lab = CONV), enclose the
        table name in percent signs (%), e.g., %tabname%.  Use the *DIM
        command to define a table.

    Notes
    -----
    Specifies surface loads on the selected lines of area regions.  The
    lines represent either the edges of area elements or axisymmetric shell
    elements themselves.  Surface loads may be transferred from lines to
    elements with the SFTRAN or SBCTRAN commands.  See the SFE command for
    a description of surface loads.  Loads input on this command may be
    tapered.  See the SFGRAD command for an alternate tapered load
    capability.

    You can specify a table name only when using structural (PRES) and
    thermal (CONV [film coefficient and/or bulk temperature], HFLUX), and
    surface emissivity and ambient temperature (RAD) surface load labels.
    VALJ and VAL2J are ignored for tabular boundary conditions.

    This command is also valid in PREP7.
    """
    command = f"SFL,{line},{lab},{vali},{valj},{val2i},{val2j}"
    return self.run(command, **kwargs)


def sfldele(self, line="", lab="", **kwargs):
    """Deletes surface loads from lines.

    APDL Command: SFLDELE

    Parameters
    ----------
    line
        Line to which surface load deletion applies.  If ALL, delete load
        from all selected lines [LSEL].  If LINE = P, graphical picking is
        enabled and all remaining command fields are ignored (valid only in
        the GUI).  A component name may be substituted for LINE.

    lab
        Valid surface load label.  If ALL, use all appropriate labels.  See
        the SFL command for labels.

    Notes
    -----
    Deletes surface loads (and all corresponding finite element loads) from
    selected lines.

    This command is also valid in PREP7.
    """
    command = f"SFLDELE,{line},{lab}"
    return self.run(command, **kwargs)


def sfllist(self, line="", lab="", **kwargs):
    """Lists the surface loads for lines.

    APDL Command: SFLLIST

    Parameters
    ----------
    line
        Line at which surface load is to be listed.  If ALL (or blank),
        list for all selected lines [LSEL].  If LINE = P, graphical picking
        is enabled and all remaining command fields are ignored (valid only
        in the GUI).  A component name may be substituted for LINE.

    lab
        Valid surface load label.  If ALL (or blank), use all appropriate
        labels.  See the SFL command for labels.

    Notes
    -----
    Lists the surface loads for the specified line.

    This command is valid in any processor.
    """
    command = f"SFLLIST,{line},{lab}"
    return self.run(command, **kwargs)


def sftran(self, **kwargs):
    """Transfer the solid model surface loads to the finite element model.

    APDL Command: SFTRAN

    Notes
    -----
    Surface loads are transferred only from selected lines and areas to all
    selected elements.  The SFTRAN operation is also done if the SBCTRAN
    command is issued or automatically done upon initiation of the solution
    calculations [SOLVE].

    This command is also valid in PREP7.
    """
    command = f"SFTRAN,"
    return self.run(command, **kwargs)



import os


def afun(self, lab="", **kwargs):
    """Specifies units for angular functions in parameter expressions.

    APDL Command: *AFUN

    Parameters
    ----------
    lab
        Specifies the units to be used:

        Use radians for input and output of parameter angular functions (default). - Use degrees for input and output of parameter angular functions.

    Notes
    -----
    Only the SIN, COS, TAN, ASIN, ACOS, ATAN, ATAN2, ANGLEK, and ANGLEN
    functions [*SET, *VFUN] are affected by this command.
    """
    command = f"*AFUN,{lab}"
    return self.run(command, **kwargs)


def dim(self, par="", type_="", imax="", jmax="", kmax="", var1="", var2="",
        var3="", csysid="", **kwargs):
    """Defines an array parameter and its dimensions.

    APDL Command: *DIM

    Parameters
    ----------
    par
        Name of parameter to be dimensioned.  See *SET for name
        restrictions.

    type\_
        Array type:

        Arrays are similar to standard FORTRAN arrays (indices are integers) (default).  Index numbers for the rows, columns, and planes are sequential values beginning with one. Used for 1-, 2-, or 3-D arrays. - Same as ARRAY, but used to specify 4-D arrays.

        Same as ARRAY, but used to specify 5-D arrays. - Array entries are character strings (up to 8 characters each).  Index numbers
                          for rows, columns, and planes are sequential
                          values beginning with one.

        Array indices are real (non-integer) numbers which must be defined when filling the table.  Index numbers for the rows and columns are stored in the zero column and row "array elements" and are initially assigned a near-zero value.  Index numbers must be in ascending order and are used only for retrieving an array element.  When retrieving an array element with a real index that does not match a specified index, linear interpolation is done among the nearest indices and the corresponding array element values [*SET]. Used for 1-, 2-, or 3-D tables. - Same as TABLE, but used to specify 4-D tables.

        Same as TABLE, but used to specify 5-D tables. - Array entries are character strings (up to IMAX each). Index numbers for
                          columns and planes are sequential values
                          beginning with 1. Row index is character position
                          in string.

    imax
        Extent of first dimension (row). (For Type = STRING, IMAX is
        rounded up to the next multiple of eight and has a limit of 248).
        Defaults to 1.

    jmax
        Extent of second dimension (column).  Defaults to 1.

    kmax
        Extent of third dimension (plane).  Defaults to 1.

    var1
        Variable name corresponding to the first dimension (row) for Type =
        TABLE.  Defaults to Row.

    var2
        Variable name corresponding to the second dimension (column) for
        Type = TABLE.  Defaults to Column.

    var3
        Variable name corresponding to the third dimension (plane) for Type
        = TABLE.  Defaults to Plane.

    csysid
        An integer corresponding to the coordinate system ID Number.

    Notes
    -----
    Up to three dimensions (row, column, and plane) may be defined using
    ARRAY and TABLE.  Use ARR4, ARR5, TAB4, and TAB5 to define up to five
    dimensions (row, column, plane, book, and shelf). An index number is
    associated with each row, column, and plane.  For array and table type
    parameters, element values are initialized to zero.  For character and
    string parameters, element values are initialized to (blank).  A
    defined parameter must be deleted [*SET] before its dimensions can be
    changed.  Scalar (single valued) parameters should not be dimensioned.
    *DIM,A,,3 defines a vector array with elements A(1), A(2), and A(3).
    *DIM,B,,2,3 defines a 2x3 array with elements B(1,1), B(2,1), B(1,2),
    B(2,2), B(1,3), and B(2,3).  Use *STATUS,Par to display elements of
    array Par. You can write formatted data files (tabular formatting) from
    data held in arrays through the *VWRITE command.

    If you use table parameters to define boundary conditions, then Var1,
    Var2, and/or Var3 can either specify a primary variable (listed in
    Table: 130:: *DIM - Primary Variables) or can be an independent
    parameter.  If specifying an independent parameter, then you must
    define an additional table for the independent parameter.  The
    additional table must have the same name as the independent parameter
    and may be a function of one or more primary variables or another
    independent parameter.  All independent parameters must relate to a
    primary variable.

    Tabular load arrays can be defined in both global Cartesian (default)
    or local (see below) coordinate systems by specifying CSYSID, as
    defined in LOCAL. For batch operations, you must specify your
    coordinate system first.

    The following constraints apply when you specify a local coordinate
    system for your tabular loads:

    If you are specifying a 4- or 5-D array or table, four additional
    fields (LMAX, MMAX, Var4, and Var5) are available. Thus, for a 4-D
    table, the command syntax would be:

    For a 5-D table, the command syntax would be:

    You cannot create or edit 4- or 5-D arrays or tables using the GUI.

    See Array Parameters for a detailed discussion on and examples for
    using array parameters.

    Table: 130:: : *DIM - Primary Variables

    Specify PRESSURE as the independent variable (not PRES).

    The X, Y, and Z coordinate locations listed above are valid in global
    Cartesian, or local (Cartesian, cylindrical and spherical) coordinate
    systems. The VELOCITY label is applicable only to the calculated fluid
    velocity in element FLUID116.

    When using PRESSURE as a primary variable, the underlying element must
    have the pressure DOF associated with it, or it must be a supported
    contact element.

    The gap/penetration label (GAP) is only used for defining certain
    contact element real constants.

    The frequency label (FREQ) is valid for harmonic analyses only.

    The OMEGS, ECCENT, and THETA primary variables only apply to the
    COMBI214 element. The amplitude of the rotational velocity (OMEGS) is
    an absolute value, so only positive values of OMEGS are valid. The
    eccentricity (ECCENT) and phase shift (THETA) labels are only valid for
    nonlinear analyses.

    If you use table parameters to define boundary conditions, the table
    names (Par) must not exceed 32 characters.

    In thermal analyses, if you apply tabular loads as a function of
    temperature but the rest of the model is linear (e.g., includes no
    temperature-dependent material properties or radiation ), you should
    turn on Newton-Raphson iterations (NROPT,FULL) to evaluate the
    temperature-dependent tabular boundary conditions correctly.

    This command is valid in any processor.
    """
    command = f"*DIM,{par},{type_},{imax},{jmax},{kmax},{var1},{var2},{var3},{csysid}"
    return self.run(command, **kwargs)


def get(self, par="", entity="", entnum="", item1="", it1num="", item2="",
        it2num="", **kwargs):
    """Retrieves a value and stores it as a scalar parameter or part of an

    APDL Command: *GET
    array parameter.

    Parameters
    ----------
    par
        The name of the resulting parameter. See *SET for name
        restrictions.

    entity
        Entity keyword. Valid keywords are NODE, ELEM, KP, LINE, AREA,
        VOLU, PDS, etc., as shown for Entity = in the tables below.

    entnum
        The number or label for the entity (as shown for ENTNUM = in the
        tables below). In some cases, a zero (or blank) ENTNUM represents
        all entities of the set.

    item1
        The name of a particular item for the given entity. Valid items are
        as shown in the Item1 columns of the tables below.

    it1num
        The number (or label) for the specified Item1 (if any). Valid
        IT1NUM values are as shown in the IT1NUM columns of the tables
        below. Some Item1 labels do not require an IT1NUM value.

    item2, it2num
        A second set of item labels and numbers to further qualify the item
        for which data are to be retrieved. Most items do not require this
        level of information.

    Notes
    -----
    *GET retrieves a value for a specified item and stores the value as a
    scalar parameter, or as a value in a user-named array parameter. An
    item is identified by various keyword, label, and number combinations.
    Usage is similar to the *SET command except that the parameter values
    are retrieved from previously input or calculated results. For example,
    *GET,A,ELEM,5,CENT,X returns the centroid x-location of element 5 and
    stores the result as parameter A. *GET command operations, along with
    the associated Get functions return values in the active coordinate
    system unless stated otherwise. A Get function is an alternative in-
    line function that can be used to retrieve a value instead of the *GET
    command (see Using In-line Get Functions for more information).

    Both *GET and *VGET retrieve information from the active data stored in
    memory. The database is often the source, and sometimes the information
    is retrieved from common memory blocks that the program uses to
    manipulate information. Although POST1 and POST26 operations use a
    *.rst file, *GET data is accessed from the database or from the common
    blocks. Get operations do not access the *.rst file directly. For
    repeated gets of sequential items, such as from a series of elements,
    see the *VGET command.

    Most items are stored in the database after they are calculated and are
    available anytime thereafter. Items are grouped according to where they
    are usually first defined or calculated. Preprocessing data will often
    not reflect the calculated values generated from section data. Do not
    use *GET to obtain data from elements that use calculated section data,
    such as beams or shells. Most of the general items listed below are
    available from all modules. Each of the sections for accessing *GET
    parameters are shown in the following order:

    *GET General Entity Items

    *GET Preprocessing Entity Items

    *GET Solution Entity Items

    *GET Postprocessing Entity Items

    *GET Probabilistic Design Entity Items

    The *GET command is valid in any processor.
    """
    command = f"*GET,{par},{entity},{entnum},{item1},{it1num},{item2},{it2num}"
    return self.run(command, **kwargs)


def inquire(self, func):
    """Returns system information.

    Parameters
    ----------
    func : str
       Specifies the type of system information returned.  See the
       notes section for more information.

    Returns
    -------
    value : str
        Value of the inquired item.

    Notes
    -----
    Allowable func entries
    LOGIN - Returns the pathname of the login directory on Linux
    systems or the pathname of the default directory (including
    drive letter) on Windows systems.

    - ``DOCU`` - Pathname of the ANSYS docu directory.
    - ``APDL`` - Pathname of the ANSYS APDL directory.
    - ``PROG`` - Pathname of the ANSYS executable directory.
    - ``AUTH`` - Pathname of the directory in which the license file resides.
    - ``USER`` - Name of the user currently logged-in.
    - ``DIRECTORY`` - Pathname of the current directory.
    - ``JOBNAME`` - Current Jobname.
    - ``RSTDIR`` - Result file directory
    - ``RSTFILE`` - Result file name
    - ``RSTEXT`` - Result file extension
    - ``OUTPUT`` - Current output file name

    Examples
    --------
    Return the job name

    >>> mapdl.inquire('JOBNAME')
    file

    Return the result file name

    >>> mapdl.inquire('RSTFILE')
    'file.rst'
    """
    response = self.run(f'/INQUIRE,,{func}', mute=False)
    if '=' in response:
        return response.split('=')[1].strip()
    return ''


def parres(self, lab="", fname="", ext="", **kwargs):
    """Reads parameters from a file.

    APDL Command: PARRES

    Parameters
    ----------
    lab
        Read operation.

        NEW - Replace current parameter set with these parameters (default).

        CHANGE - Extend current parameter set with these
        parameters, replacing any that already exist.

    fname
        File name and directory path (248 characters maximum,
        including the characters needed for the directory path).
        An unspecified directory path defaults to the working
        directory; in this case, you can use all 248 characters
        for the file name.

        The file name defaults to Jobname.

    ext
        Filename extension (eight-character maximum).  The
        extension defaults to PARM if Fname is blank.

    Examples
    --------
    Read a local parameter file.

    >>> mapdl.parres('parm.PARM')

    Notes
    -----
    Reads parameters from a coded file.  The parameter file may
    have been written with the PARSAV command.  The parameters
    read may replace or change the current parameter set.

    This command is valid in any processor.
    """
    if ext:
        fname = fname + '.' + ext
    elif not fname:
        fname = '.' + 'PARM'

    if 'Grpc' in self.__class__.__name__:  # grpc mode
        if self._local:
            if not os.path.isfile(fname):
                raise FileNotFoundError(
                    'Unable to locate filename "%s"' % fname)

            if not os.path.dirname(fname):
                filename = os.path.join(os.getcwd(), fname)
            else:
                filename = fname
        else:
            if not os.path.dirname(fname):
                # might be trying to run a local file.  Check if the
                # file exists remotely.
                if fname not in self.list_files():
                    self.upload(fname, progress_bar=False)
            else:
                self.upload(fname, progress_bar=False)
            filename = os.path.basename(fname)
    else:
        filename = fname

    return self.input(filename)


def parsav(self, lab="", fname="", ext="", **kwargs):
    """Writes parameters to a file.

    APDL Command: PARSAV

    Parameters
    ----------
    lab
        Write operation:

        - ``'SCALAR'`` : Write only scalar parameters (default). 
        - ``'ALL'`` : Write scalar and array parameters.
          Parameters may be numeric or alphanumeric.

    fname
        File name and directory path (248 characters maximum,
        including the characters needed for the directory path).
        An unspecified directory path defaults to the working
        directory; in this case, you can use all 248 characters
        for the file name.

    ext
        Filename extension (eight-character maximum).

    Notes
    -----
    Writes the current parameters to a coded file.  Previous
    parameters on this file, if any, will be overwritten.  The
    parameter file may be read with the PARRES command.

    PARSAV/PARRES operations truncate some long decimal strings,
    and can cause differing values in your solution data when
    other operations are performed. A good practice is to limit
    the number of decimal places you will use before and after
    these operations.

    This command is valid in any processor.
    """
    command = f"PARSAV,{lab},{fname},{ext}"
    return self.run(command, **kwargs)


def starset(self, par="", value="", val2="", val3="", val4="", val5="",
            val6="", val7="", val8="", val9="", val10="", **kwargs):
    """Assigns values to user-named parameters.

    APDL Command: *SET

    Parameters
    ----------
    par
        An alphanumeric name used to identify this parameter.  Par may be
        up to 32 characters, beginning with a letter and containing only
        letters, numbers, and underscores.  Examples:  ABC   A3X   TOP_END.
        ANSYS command names, function names, label names, component and
        assembly names, etc., should not be used.  Parameter names
        beginning with an underscore (e.g.,  _LOOP) are reserved for use by
        ANSYS and should be avoided.  Parameter names ending in an
        underscore are not listed by the *STATUS command.  Array parameter
        names must be followed by a subscript, and the entire expression
        must be 32 characters or less.  Examples:  A(1,1)   NEW_VAL(3,2,5)
        RESULT(1000).  There is no character parameter substitution for the
        Par field. Table parameters that are used in command fields (where
        constant values are normally given) are limited to 32 characters.

    value
        Numerical value or alphanumeric character string (up to 32
        characters enclosed in single quotes) to be assigned to this
        parameter.  Examples:  A(1,3)=7.4 B='ABC3'.  May also be a
        parameter or a parametric expression.  Examples:  C=A(1,3)
        A(2,2)=(C+4)/2.  If blank, delete this parameter.  Example:  A=
        deletes parameter A.

    val2, val3, val4, val5, val6, val7, val8, val9, val10
        If Par is an array parameter, values VAL2 through VAL10 (up to the
        last nonblank value) are sequentially assigned to the succeeding
        array elements of the column.  Example:  *SET,A(1,4),10,11 assigns
        A(1,4)=10, A(2,4)=11.  *SET,B(2,3),'file10','file11' assigns
        B(2,3)='file10', B(3,3)='file11'.

    Notes
    -----
    Assigns values to user-named parameters that may be substituted later
    in the run.  The equivalent (and recommended) format is

    Par = VALUE,VAL2,VAL3, . . . , VAL10

    which may be used in place of  *SET,Par, : ... for convenience.

    This command is valid in any processor.

    Parameters (numeric or character) may be scalars (single valued) or
    arrays (multiple valued in one, two, or three dimensions). An unlimited
    number of parameter names may be defined in any ANSYS run. For very
    large numbers of parameters, it is most efficient to define them in
    alphabetical order.

    Parameter values may be redefined at any time.  Array parameters may
    also be assigned values within a do-loop [*DO] for convenience.
    Internally programmed do-loop commands are also available with the *VXX
    commands (*VFILL).  Parameter values (except for parameters ending in
    an underscore) may be listed with the  *STATUS command, displayed with
    the *VPLOT   command (numeric parameters only), and modified with the
    *VEDIT command (numeric parameters only).

    Older ANSYS-supplied macro files may use parameter names that do not
    begin with an underscore. Using these macros embedded in your own
    macros may cause conflicts if the same parameter names are used.

    Parameters can also be resolved in comments created by the /COM command
    (see /COM for complete documentation). A parameter can be deleted by
    redefining it with a blank  VALUE.  If the parameter is an array, the
    entire array is deleted.  Parameters may also be defined by a response
    to a query with the  *ASK command or from an "ANSYS-supplied" value
    with the *GET command.

    Array parameters must be dimensioned  [*DIM] before being assigned
    values unless they are the result of an array operation or defined
    using the implied loop convention. Scalar parameters that are not
    defined are initialized to a "near" zero value.  Numeric array
    parameters are initialized to zero when dimensioned, and character
    array parameters are initialized to blank.  An existing array parameter
    must be deleted before it can be redimensioned.  Array parameter names
    must be followed by a subscript list (enclosed in parentheses)
    identifying the element of the array.  The subscript list may have one,
    two, or three values (separated by commas).  Typical array parameter
    elements are  A(1,1), NEW_VAL(3,2,5), RESULT(1000).  Subscripts for
    defining an array element must be integers (or parameter expressions
    that evaluate to integers).  Non-integer values are rounded to the
    nearest integer value.  All array parameters are stored as 3-D arrays
    with the unspecified dimensions set to 1.  For example, the 4th array
    element of a 1-dimensional array, A(4), is stored as array element
    A(4,1,1).  Arrays are patterned after standard FORTRAN conventions.

    If the parameter name Par is input in a numeric argument of a command,
    the numeric value of the parameter (as assigned with *SET, *GET, =,
    etc.) is substituted into the command at that point.  Substitution
    occurs only if the parameter name is used between blanks, commas,
    parentheses, or arithmetic operators (or any combination) in a numeric
    argument.  Substitution can be prevented by enclosing the parameter
    name Par within single quotes ( ' ), if the parameter is alone in the
    argument; if the parameter is part of an arithmetic expression, the
    entire expression must be enclosed within single quotes to prevent
    substitution.  In either case the character string will be used instead
    of the numeric value (and the string will be taken as 0.0 if it is in a
    numeric argument).

    A forced substitution is available in the text fields of the /TITLE,
    /STITLE,  /TLABEL, /AN3D, /SYP (ARG1--ARG8), and *ABBR  commands by
    enclosing the parameter within percent (%) signs.  Also, parameter
    substitution may be forced within the file name or extension fields of
    commands having these fields by enclosing the parameter within percent
    (%) signs.  Array parameters  [*DIM] must include a subscript (within
    parentheses) to identify the array element whose value is to be
    substituted, such as A(1,3).  Out-of-range subscripts result in an
    error message.  Non-integer subscripts are allowed when identifying a
    TABLE array element for substitution.  A proportional linear
    interpolation of values among the nearest array elements is performed
    before substitution.  Interpolation is done in all three dimensions.

    Note:: : Interpolation is based upon the assigned index numbers which
    must be defined when the table is filled [*DIM].

    Most alphanumeric arguments permit the use of character parameter
    substitution.  When the parameter name Par input, the alphanumeric
    value of the parameter is substituted into the command at that point.
    Substitution can be suppressed by enclosing the parameter name within
    single quotes ( ' ).  Forced substitution is available in some fields
    by enclosing the parameter name within percent (%) signs.  Valid forced
    substitution fields include command name fields, Fname (filename) or
    Ext (extension) arguments, *ABBR command (Abbr arguments), /TITLE and
    /STITLE commands (Title argument) and /TLABEL command (Text argument).
    Character parameter substitution is also available in the  *ASK, /AN3D,
    *CFWRITE,  *IF,  *ELSEIF,   *MSG,  *SET,  *USE,  *VREAD, and  *VWRITE
    commands.   Character array parameters must include a subscript (within
    parentheses) to identify the array element whose value is to be
    substituted.

    If a parameter operation expression is input in a numeric argument, the
    numeric value of the expression is substituted into the command at that
    point.  Allowable operation expressions are of the form

    E1oE2oE3: ...oE10

    where E1, E2, etc. are expressions connected by operators (o).  The
    allowable operations (o) are

    ``+ - * / ** < >``

    For example, ``A+B**C/D*E`` is a valid operation expression.  The ``*``
    represents multiplication and the ``**`` represents exponentiation.
    """
    command = f"*SET,{par},{value},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9},{val10}"
    return self.run(command, **kwargs)


def starvget(self, parr="", entity="", entnum="", item1="", it1num="",
             item2="", it2num="", kloop="", **kwargs):
    """Retrieves values and stores them into an array parameter.

    APDL Command: *VGET

    Parameters
    ----------
    parr
        The name of the resulting vector array parameter.  See *SET for
        name restrictions.

    entity
        Entity keyword.  Valid keywords are NODE, ELEM, KP, LINE, AREA,
        VOLU, etc. as shown for Entity = in the tables below.

    entnum
        The number of the entity (as shown for ENTNUM = in the tables
        below).

    item1
        The name of a particular item for the given entity.  Valid items
        are as shown in the Item1 columns of the tables below.

    it1num
        The number (or label) for the specified Item1 (if any).  Valid
        IT1NUM values are as shown in the IT1NUM columns of the tables
        below.  Some Item1 labels do not require an IT1NUM value.

    item2, it2num
        A second set of item labels and numbers to further qualify the item
        for which data is to be retrieved.  Most items do not require this
        level of information.

    kloop
        Field to be looped on:

        ``0`` or ``2`` : Loop on the ENTNUM field (default).

        ``3`` : Loop on the Item1 field.

        ``4`` : Loop on the IT1NUM field. Successive items are as shown with IT1NUM.

        ``5`` : Loop on the Item2 field.

        ``6`` : Loop on the IT2NUM field. Successive items are as shown with IT2NUM.

    Notes
    -----
    Retrieves values for specified items and stores the values in an output
    vector of a user-named array parameter according to:

    ``ParR = f(Entity, ENTNUM, Item1, IT1NUM, Item2, IT2NUM)``

    where (f) is the *GET function; Entity, Item1, and Item2 are keywords;
    and ENTNUM, IT1NUM, and IT2NUM are numbers or labels corresponding to
    the keywords. Looping continues over successive entity numbers (ENTNUM)
    for the KLOOP default.  For example, *VGET,A(1),ELEM,5,CENT,X returns
    the centroid x-location of element 5 and stores the result in the first
    location of A.  Retrieving continues with element 6, 7, 8, etc.,
    regardless of whether the element exists or is selected, until
    successive array locations are filled.  Use *VLEN or *VMASK to skip
    locations. Absolute values and scale factors may be applied to the
    result parameter [*VABS, *VFACT].  Results may be cumulative [*VCUM].
    See the *VOPER command for general details.  Results can be put back
    into an analysis by writing a file of the desired input commands with
    the *VWRITE command.  See also the *VPUT command.

    Both *GET and *VGET retrieve information from the active data stored in
    memory. The database is often the source, and sometimes the information
    is retrieved from common memory blocks that ANSYS uses to manipulate
    information. Although POST1 and POST26 operations use a *.rst file, GET
    data is accessed from the database or from the common blocks. Get
    operations do not access the *.rst file directly.

    The *VGET command retrieves both the unprocessed real and the imaginary
    parts (original and duplicate sector nodes and elements) of a cyclic
    symmetry solution.

    For each  of the sections for accessing *VGET parameters see:
    https://www.mm.bme.hu/~gyebro/files/ans_help_v182/ans_cmd/Hlp_C_VGET_st.html

    This command is valid in any processor.
    """
    command = f"*VGET,{parr},{entity},{entnum},{item1},{it1num},{item2},{it2num},{kloop}"
    return self.run(command, **kwargs)


def taxis(self, parmloc="", naxis="", val1="", val2="", val3="", val4="",
          val5="", val6="", val7="", val8="", val9="", val10="", **kwargs):
    """Defines table index numbers.

    APDL Command: *TAXIS

    Parameters
    ----------
    parmloc
        Name and starting location in the table array parameter for
        indexing. Indexing occurs along the axis defined with nAxis.

    naxis
        Axis along which indexing occurs.  Valid labels are:

        Corresponds to Row. Default. - Corresponds to Column.

        Corresponds to Plane. - Corresponds to Book.

        Corresponds to Shelf. - Lists all index numbers. Valid only if Val1 = LIST.

    val1, val2, val3, . . . , val10
        Values of the index numbers for the axis nAxis, starting from the
        table array parameter location ParmLoc. You can define up to ten
        values.

    Notes
    -----
    *TAXIS is a convenient method to define table index values. These
    values reside in the zero column, row, etc. Instead of filling values
    in these zero location spots, use the *TAXIS command. For example,

     would fill index values 1.0, 2.2, 3.5, 4.7, and 5.9 in nAxis 2 (column
    location), starting at location 4.

    To list index numbers, issue *TAXIS,ParmLoc, nAxis, LIST, where nAxis =
    1 through 5 or ALL.
    """
    command = f"*TAXIS,{parmloc},{naxis},{val1},{val2},{val3},{val4},{val5},{val6},{val7},{val8},{val9},{val10}"
    return self.run(command, **kwargs)


def tread(self, par="", fname="", ext="", nskip="", **kwargs):
    """Reads data from an external file into a table array parameter.

    APDL Command: *TREAD

    Parameters
    ----------
    par
        Table array parameter name as defined by the *DIM command.

    fname
        File name and directory path (248 characters maximum, including the
        characters needed for the directory path).  An unspecified
        directory path defaults to the working directory; in this case, you
        can use all 248 characters for the file name.

    ext
        Filename extension (eight-character maximum).

    nskip
        Number of comment lines at the beginning of the file being read
        that will be skipped during the reading.  Default = 0.

    Notes
    -----
    Use this command to read in a table of data from an external file into
    an ANSYS table array parameter.  The external file may be created using
    a text editor or by an external application or program.  The external
    file must be in tab-delimited, blank-delimited, or comma-delimited
    format to be used by *TREAD. The ANSYS TABLE type array parameter must
    be defined before you can read in an external file.  See *DIM  for more
    information.

    This command is not applicable to 4- or 5-D tables.
    """
    command = f"*TREAD,{par},{fname},{ext},,{nskip}"
    return self.run(command, **kwargs)


def vfill(self, parr="", func="", con1="", con2="", con3="", con4="",
          con5="", con6="", con7="", con8="", con9="", con10="", **kwargs):
    """Fills an array parameter.

    APDL Command: *VFILL

    Parameters
    ----------
    parr
        The name of the resulting numeric array parameter vector.  See *SET
        for name restrictions.

    func
        Fill function:

        DATA - Assign specified values CON1, CON2, etc. to successive
        array elements.  Up to 10 assignments may be made at a
        time.  Any CON values after a blank CON value are
        ignored. - Assign ramp function values: CON1+((n-1)*CON2)
        , where n is the loop number [*VLEN].  To specify a
        constant function (no ramp), set CON2 to zero.

        RAMP - Assign random number values based on a uniform
        distribution RAND(CON1,CON2), where: - Assign random
        sample of Gaussian distributions GDIS(CON1,CON2).

        RAND - Assign random number values based on a uniform
        distribution RAND(CON1,CON2), where CON1 is the lower
        bound (defaults to 0.0) and CON2 is the upper bound
        (defaults to 1.0)

        GDIS - Assign random sample of Gaussian distributions
        GDIS(CON1,CON2) where CON1 is the mean (defaults to 0.0),
        and CON2 is the standard deviation (defaults to 1.0)

        TRIA - Assigns random number values based on a triangular
        distribution TRIA(CON1,CON2,CON3) where CON1 is the lower
        bound (defaults to 0.0), CON2 is the location of the peak
        value (CON1 ≤ CON2 ≤CON3; CON2 defaults to 0 if CON1 ≤ 0 ≤
        CON3, CON1 if 0 ≤ CON1, or CON3 if CON3 ≤ 0), and CON3 is
        the upper bound (defaults to 1.0 + CON1 if CON1 ≥ 0 or 0.0
        if CON1 ≤ 0)

        BETA - Assigns random number values based on a beta
        distribution BETA(CON1,CON2,CON3,CON4) where: CON1 is the
        lower bound (defaults to 0.0), CON2 is the upper bound
        (defaults to 1.0 + CON1 if CON1 ≥ 0 or 0.0 if CON1 ≤ 0), and CON3
        and CON4 are the alpha and beta parameters, respectively,
        of the beta function. Alpha and beta must both be
        positive; they default to 1.0.

        GAMM - Assigns random number values based on a gamma
        distribution: GAMM(CON1,CON2,CON3) where: CON1 is the
        lower bound (defaults to 0.0), CON2 and CON3 are the alpha
        and beta parameters, respectively, of the gamma
        function. Alpha and beta must both be positive; they
        default to 1.0.

        RIGID - Generates the rigid body modes with respect to the
        reference point coordinates (CON1, CON2, CON3). The
        dimensions of the array parameter ParR are (dim1,dim2)
        where dim1 is the maximum node number (including internal
        nodes but excluding orientation nodes) multiplied by the
        number of degrees of freedom, and dim2 is the number of
        rigid body modes (which corresponds to the number of
        structural degrees of freedom).

        CLUSTER - Generates excitation frequencies with clustering
        option CLUSTER(CON1,CON2,CON3,CON4,%CON5%) where:

        - CON1 is the lower end of the frequency range in Hz (0 < CON1)
        - CON2 is the upper end of the frequency range in Hz (CON1 < CON2)
        - CON3 is the number of points on each side of the natural
          frequency (4 ≤ CON3 ≤ 20, defaults to 4)
        - CON4 is the constant damping ratio value or an array
          parameter (size NFR) specifying the damping ratios (if
          zero or blank, defaults to constant damping ratio of
          0.005)
        - CON5 is an array parameter (size NFR) specifying the
          natural frequencies in Hz

        The dimension of the resulting array parameter ParR is
        less than 2+NFR*(2*CON3+1) where NFR is the number of
        natural frequencies defined in CON5.

    con1, con2, con3, . . . , con10
        Constants used with above functions.

    Notes
    -----
    Operates on input data and produces one output array parameter vector
    according to:

    ParR = f(CON1, CON2, : ...)

    where the functions (f) are described above. Operations use successive
    array elements [*VLEN, *VMASK] with the default being all successive
    elements.  For example, *VFILL,A,RAMP,1,10 assigns A(1) = 1.0, A(2) =
    11.0, A(3) = 21.0, etc.  *VFILL,B(5,1),DATA,1.5,3.0 assigns B(5,1) =
    1.5 and B(6,1) = 3.0.  Absolute values and scale factors may be applied
    to the result parameter [*VABS, *VFACT].  Results may be cumulative
    [*VCUM].  See the *VOPER command for details.

    This command is valid in any processor.
    """
    command = f"*VFILL,{parr},{func},{con1},{con2},{con3},{con4},{con5},{con6},{con7},{con8},{con9},{con10}"
    return self.run(command, **kwargs)


def vread(self, parr="", fname="", ext="", label="", n1="", n2="", n3="",
          nskip="", **kwargs):
    """Reads data and produces an array parameter vector or matrix.

    APDL Command: *VREAD

    Parameters
    ----------
    parr
        The name of the resulting array parameter vector.  See *SET for
        name restrictions.  The parameter must exist as a dimensioned array
        [*DIM]. String arrays are limited to a maximum of 8 characters.

    fname
        File name and directory path (248 characters maximum, including the
        characters needed for the directory path).  An unspecified
        directory path defaults to the working directory; in this case, you
        can use all 248 characters for the file name.

    ext
        Filename extension (eight-character maximum).

    label
        Can take a value of IJK, IKJ, JIK, JKI, KIJ, KJI, or blank (IJK).

    n1, n2, n3
        Read as (((ParR (i,j,k), k = 1,n1), i = 1, n2), j = 1, n3) for
        Label = KIJ. n2 and n3 default to 1.

    nskip
        Number of lines at the beginning of the file being read that will
        be skipped during the reading.  Default = 0.

    Notes
    -----
    Reads data from a file and fills in an array parameter vector or
    matrix.  Data are read from a formatted file or, if the menu is off
    [/MENU,OFF] and Fname is blank, from the next input lines.  The format
    of the data to be read must be input immediately following the *VREAD
    command.  The format specifies the number of fields to be read per
    record, the field width, and the placement of the decimal point (if
    none specified in the value).  The read operation follows the available
    FORTRAN FORMAT conventions of the system (see your system FORTRAN
    manual).  Any standard FORTRAN real format (such as (4F6.0),
    (E10.3,2X,D8.2), etc.) or alphanumeric format (A) may be used.
    Alphanumeric strings are limited to a maximum of 8 characters for any
    field (A8). For storage of string arrays greater than 8 characters, the
    *SREAD command can be used. Integer (I) and list-directed (*)
    descriptors may not be used.  The parentheses must be included in the
    format and the format must not exceed 80 characters (including
    parentheses).  The input line length is limited to 128 characters.

    A starting array element number must be defined for the result array
    parameter vector (numeric or character).  For example, entering these
    two lines:

    will read two values from each line of file ARRAYVAL and assign the
    values to A(1), A(2), A(3), etc.  Reading continues until successive
    row elements [*VLEN, *VMASK, *DIM] are filled.

    For an array parameter matrix, a starting array element row and column
    number must be defined.  For example, entering these two lines:

    will read two values from each line of file ARRAYVAL and assign the
    values to A(1,1), A(2,1), A(3,1), etc.  Reading continues until n1 (10)
    successive row elements are filled.  Once the maximum row number is
    reached, subsequent data will be read into the next column (e.g.,
    A(1,2), A(2,2), A(3,2), etc.)

    For numerical parameters, absolute values and scale factors may be
    applied to the result parameter [*VABS, *VFACT].  Results may be
    cumulative [*VCUM].  See the *VOPER command for details.  If you are in
    the GUI the *VREAD command must be contained in an externally prepared
    file read into the ANSYS program (i.e., *USE, /INPUT, etc.).

    This command is not applicable to 4- or 5-D arrays.

    This command is valid in any processor.
    """
    command = f"*VREAD,{parr},{fname},{ext},,{label},{n1},{n2},{n3},{nskip}"
    return self.run(command, **kwargs)



def hptcreate(self, type_="", entity="", nhp="", label="", val1="", val2="",
              val3="", **kwargs):
    """Defines a hard point.

    APDL Command: HPTCREATE

    Parameters
    ----------
    type\_
        Type of entity on which the hard point will be created.

        LINE - Hard point will be created on a line.

        AREA - Hard point will be created within an area (not on the boundaries).

    entity
        Number of the line or area on which the hard point will be created.

    nhp
        Number assigned to the hard point.  Defaults to the lowest
        available hard point number.

    label
        If LABEL = COORD, VAL1, VAL2, and VAL3 are the respective global X,
        Y, and Z coordinates.  If LABEL = RATIO, VAL1 is the parameter
        value (this is available only for lines).  Valid parameter values
        are between 0 and 1.  VAL2 and VAL3 are ignored.

    val1
        If LABEL = RATIO, ratio value for line.  If LABEL = COORD, global X
        coordinate value.

    val2
        If LABEL = COORD, global Y coordinate value.

    val3
        If LABEL = COORD, global Z coordinate value.

    Notes
    -----
    The ability to enter a parameter value provides a simple way of
    positioning hard points on lines.  For example, to place a hard point
    halfway along a line, one can simply specify a VAL1 value of 0.5.

    For models imported through the DEFAULT IGES filter, you can place hard
    points on models only by specifying coordinates (you can't place a hard
    point using interactive picking).

    If you issue any commands that update the geometry of an entity, such
    as Boolean or simplification commands, any hard points associated with
    that entity are deleted.  Therefore, you should add any hard points
    after completing the solid model.  If you delete an entity that has
    associated hard points, those hard points are either

    Deleted along with the entity (if the hard point is not associated with
    any other entities).

    Detached from the deleted entity (if the hard point is associated with
    additional entities).

    When archiving your model (CDWRITE), hardpoint information cannot be
    written to the IGES file. The Jobname.cdb file can be written with the
    CDWRITE,DB option.

    Hard points are only applicable for area and volume meshing, not for
    beams.
    """
    command = f"HPTCREATE,{type_},{entity},{nhp},{label},{val1},{val2},{val3}"
    return self.run(command, **kwargs)


def hptdelete(self, np1="", np2="", ninc="", **kwargs):
    """Deletes selected hardpoints.

    APDL Command: HPTDELETE

    Parameters
    ----------
    np1, np2, ninc
        Delete the pattern of hard points beginning with NP1 to NP2 in
        steps of NINC (defaults to 1).  If NP1 = ALL, NP2 and NINC are
        ignored and the pattern is all selected hard points [KSEL].  If NP1
        = P, graphical picking is enabled and all remaining command fields
        are ignored.

    Notes
    -----
    Deletes all attributes attached to the designated hard points as well
    as the hard points themselves.  If any entity is attached to a
    designated hard point, the command detaches the hard point from that
    entity (the program will alert you that this will occur).
    """
    command = f"HPTDELETE,{np1},{np2},{ninc}"
    return self.run(command, **kwargs)



def ce(self, neqn="", const="", node1="", lab1="", c1="", node2="",
       lab2="", c2="", node3="", lab3="", c3="", **kwargs):
    """Defines a constraint equation relating degrees of freedom.

    APDL Command: CE

    Parameters
    ----------
    neqn
        Set equation reference number:

        n - Arbitrary set number.

        HIGH - The highest defined constraint equation number. This option is especially
               useful when adding nodes to an existing set.

        NEXT - The highest defined constraint equation number plus one. This option
               automatically numbers coupled sets so that existing sets are
               not modified.

    const
        Constant term of equation.

    node1
        Node for first term of equation.  If -NODE1, this term is deleted
        from the equation.

    lab1
        Degree of freedom label for first term of equation.  Structural
        labels:  UX, UY, or UZ (displacements); ROTX, ROTY, or ROTZ
        (rotations, in radians).  Thermal labels: TEMP, TBOT, TE2, TE3, . .
        ., TTOP (temperature).  Electric labels:  VOLT (voltage).  Magnetic
        labels:  MAG (scalar magnetic potential); AX, AY, or AZ (vector
        magnetic potentials). Diffusion label: CONC (concentration).

    c1
        Coefficient for first node term of equation.  If zero, this term is
        ignored.

    node2, lab2, c2
        Node, label, and coefficient for second term.

    node3, lab3, c3
        Node, label, and coefficient for third term.

    Notes
    -----
    Repeat the CE command to add additional terms to the same equation.  To
    change only the constant term, repeat the command with no node terms
    specified.  Only the constant term can be changed during solution, and
    only with the CECMOD command.

    Linear constraint equations may be used to relate the degrees of
    freedom of selected nodes in a more general manner than described for
    nodal coupling [CP].  The constraint equation is of the form:

    where U(I) is the degree of freedom (displacement, temperature, etc.)
    of term (I).  The following example is a set of two constraint
    equations, each containing three terms:

    0.0 = 3.0* (1 UX) + 3.0* (4 UX) + (-2.0)* (4 ROTY)

    2.0 = 6.0* (2 UX) + 10.0* (4 UY) + 1.0* (3 UZ)

    The first unique degree of freedom in the equation is eliminated in
    terms of all other degrees of freedom in the equation.  A unique degree
    of freedom is one which is not specified in any other constraint
    equation, coupled node set, specified displacement set, or master
    degree of freedom set.  It is recommended that the first term of the
    equation be the degree of freedom to be eliminated.  The first term of
    the equation cannot contain a master degree of freedom, and no term can
    contain coupled degrees of freedom.  The same degree of freedom may be
    specified in more than one equation but care must be taken to avoid
    over-specification (over-constraint).

    The degrees of freedom specified in the equation (i.e., UX, UY, ROTZ,
    etc.) must also be included in the model (as determined from the
    element types [ET]).  Also, each node in the equation must be defined
    on an element (any element type containing that degree of freedom will
    do).

    For buckling and modal analyses, the constant term of the equation will
    not be taken into account (that is, CONST is always zero).

    Note that under certain circumstances a constraint equation generated
    by CE may be modified during the solution. See Program Modification of
    Constraint Equations for more information.
    """
    command = f"CE,{neqn},{const},{node1},{lab1},{c1},{node2},{lab2},{c2},{node3},{lab3},{c3}"
    return self.run(command, **kwargs)


def cecyc(self, lowname="", highname="", nsector="", hindex="",
          tolerance="", kmove="", kpairs="", **kwargs):
    """Generates the constraint equations for a cyclic symmetry analysis

    APDL Command: CECYC

    Parameters
    ----------
    lowname
        Name of a component for the nodes on the low angle edge of the
        sector.  Enclosed in single quotes.

    highname
        Name of a component for the nodes on the high angle edge of the
        sector.  Enclosed in single quotes.

    nsector
        Number of sectors in the complete 360 degrees.

    hindex
        Harmonic index to be represented by this set of constraint
        equations.  If Hindex is -1, generate constraint equations for
        static cyclic symmetry.  If HIndex is -2, generate constraint
        equations for static cyclic asymmetry.

    tolerance
        A positive tolerance is an absolute tolerance (length units), and a
        negative tolerance is a tolerance relative to the local element
        size.

    kmove
        0

        0 - Nodes are not moved.

        1 - HIGHNAME component nodes are moved to match LOWNAME component nodes exactly.

    kpairs
        0

        0 - Do not print paired nodes

        1 - Print table of paired nodes

    Notes
    -----
    The analysis can be either modal cyclic symmetry or static cyclic
    symmetry.

    The pair of nodes for which constraint equations are written are
    rotated into CSYS,1.
    """
    command = f"CECYC,{lowname},{highname},{nsector},{hindex},{tolerance},{kmove},{kpairs}"
    return self.run(command, **kwargs)


def cedele(self, neqn1="", neqn2="", ninc="", nsel="", **kwargs):
    """Deletes constraint equations.

    APDL Command: CEDELE

    Parameters
    ----------
    neqn1, neqn2, ninc
        Delete constraint equations from NEQN1 to NEQN2 (defaults to NEQN1)
        in steps of NINC (defaults to 1).  If NEQN1 = ALL, NEQN2  and NINC
        will be ignored all constraint equations will be deleted.

    nsel
        Additional node selection control:

        ANY  - Delete equation set if any of the selected nodes are in the set (default).

        ALL  - Delete equation set only if all of the selected nodes are in the set.
    """
    command = f"CEDELE,{neqn1},{neqn2},{ninc},{nsel}"
    return self.run(command, **kwargs)


def ceintf(self, toler="", dof1="", dof2="", dof3="", dof4="", dof5="",
           dof6="", movetol="", **kwargs):
    """Generates constraint equations at an interface.

    APDL Command: CEINTF

    Parameters
    ----------
    toler
        Tolerance about selected elements, based on a fraction of the
        element dimension (defaults to 0.25 (25%)).  Nodes outside the
        element by more than the tolerance are not accepted as being on the
        interface.

    dof1, dof2, dof3, . . . , dof6
        Degrees of freedom for which constraint equations are written.
        Defaults to all applicable DOFs.  DOF1 accepts ALL as a valid
        label, in which case the rest are ignored (all DOFs are applied).

    movetol
        The allowed "motion" of a node (see Note below).  This distance is
        in terms of the element coordinates (-1.0 to 1.0).  A typical value
        is 0.05.  Defaults to 0 (do not move).  MoveTol must be less than
        or equal to TOLER.

    Notes
    -----
    This command can be used to "tie" together two regions with dissimilar
    mesh patterns by generating constraint equations that connect the
    selected nodes of one region to the selected elements of the other
    region.  At the interface between regions, nodes should be selected
    from the more dense mesh region, A, and the elements selected from the
    less dense mesh region, B.  The degrees of freedom of region A nodes
    are interpolated with the corresponding degrees of freedom of the nodes
    on the region B elements, using the shape functions of the region B
    elements.  Constraint equations are then written that relate region A
    and B nodes at the interface.

    The MoveTol field lets the nodes in the previously mentioned region A
    change coordinates when slightly inside or outside the elements of
    region B.  The change in coordinates causes the nodes of region A to
    assume the same surface as the nodes associated with the elements of
    region B.   The constraint equations that relate the nodes at both
    regions of the interface are then written.

    Solid elements with six degrees of freedom should only be interfaced
    with other six degree-of-freedom elements.  The region A nodes should
    be near the region B elements.   A location tolerance based on the
    smallest region B element length may be input.  Stresses across the
    interface are not necessarily continuous.  Nodes in the interface
    region should not have specified constraints.

    Use the CPINTF command to connect nodes by coupling instead of
    constraint equations.  Use the EINTF command to connect nodes by line
    elements.  See also the NSEL and ESEL commands for selecting nodes and
    elements.  See the Mechanical APDL Theory Reference for a description
    of 3-D space used to determine if a node will be considered by this
    command.

    As an alternative to the CEINTF command, you can use contact elements
    and the internal multipoint constraint (MPC) algorithm to tie together
    two regions having dissimilar meshes. See Solid-Solid and Shell-Shell
    Assemblies for more information.
    """
    command = f"CEINTF,{toler},{dof1},{dof2},{dof3},{dof4},{dof5},{dof6},{movetol}"
    return self.run(command, **kwargs)


def celist(self, neqn1="", neqn2="", ninc="", option="", **kwargs):
    """Lists the constraint equations.

    APDL Command: CELIST

    Parameters
    ----------
    neqn1, neqn2, ninc
        List constraint equations from NEQN1 to NEQN2 (defaults to NEQN1)
        in steps of NINC (defaults to 1).  If NEQN1 = ALL (default), NEQN2
        and NINC are ignored and all constraint equations are listed.

    option
        Options for listing constraint equations:

        ANY - List equation set if any of the selected nodes are in the set (default). Only
              externally-generated constraint equations are listed.

        ALL - List equation set only if all of the selected nodes are in the set. Only
              externally-generated constraint equations are listed.

        INTE - List internally-generated constraint equations that are associated with MPC-
               based contact. Constraint equations are listed only if all
               the nodes in the set are selected.

        CONV - Convert internal constraint equations to external constraint equations.
               Internal constraint equations are converted only if all of
               the nodes in the set are selected.

    Notes
    -----
    This command is valid in any processor. However, the INTE and CONV
    options are only valid in the Solution processor after a SOLVE command
    has been issued.
    """
    command = f"CELIST,{neqn1},{neqn2},{ninc},{option}"
    return self.run(command, **kwargs)


def cerig(self, maste="", slave="", ldof="", ldof2="", ldof3="", ldof4="",
          ldof5="", **kwargs):
    """Defines a rigid region.

    APDL Command: CERIG

    Parameters
    ----------
    maste
        Retained (or master) node for this rigid region.  If MASTE = P,
        then graphical picking of the master and slave nodes is enabled
        (first node picked will be the master node, and subsequent nodes
        picked will be slave  nodes), and subsequent fields are ignored
        (valid only in GUI).

    slave
        Removed (or slave) node for this rigid region.  If ALL, slave nodes
        are all selected nodes.

    ldof
        Degrees of freedom associated with equations:

        ALL - All applicable degrees of freedom (default).  If 3-D, generate 6 equations
              based on UX, UY, UZ, ROTX, ROTY, ROTZ;  if 2-D, generate 3
              equations based on UX, UY, ROTZ.

        UXYZ - Translational degrees of freedom.  If 3-D, generate 3 equations based on the
               slave nodes' UX, UY, and UZ DOFs and the master node's UX,
               UY, UZ, ROTX, ROTY, and ROTZ DOFs;  if 2-D, generate 2
               equations based on the slave nodes UX and UY DOFs and the
               master nodes UX, UY, and ROTZ DOFs.  No equations are
               generated for the rotational coupling.

        RXYZ - Rotational degrees of freedom.  If 3-D, generate 3 equations based on ROTX,
               ROTY, ROTZ;  if 2-D, generate 1 equation based on ROTZ.  No
               equations are generated for the translational coupling.

        UX - Slave translational UX degree of freedom only.

        UY - Slave translational UY degree of freedom only.

        UZ - Slave translational UZ degree of freedom only.

        ROTX - Slave rotational ROTX degree of freedom only.

        ROTY - Slave rotational ROTY degree of freedom only.

        ROTZ - Slave rotational ROTZ degree of freedom only.

    ldof2, ldof3, ldof4, ldof5
        Additional degrees of freedom.  Used only if more than one degree
        of freedom required and Ldof is not ALL, UXYZ, or RXYZ.

    Notes
    -----
    Defines a rigid region (link, area or volume) by automatically
    generating constraint equations to relate nodes in the region.  Nodes
    in the rigid region must be assigned a geometric location before this
    command is used.  Also, nodes must be connected to elements having the
    required degree of freedom set (see Ldof above).  Generated constraint
    equations are based on small deflection theory.  Generated constraint
    equations are numbered beginning from the highest previously defined
    equation number (NEQN) plus 1.  Equations, once generated, may be
    listed [CELIST] or modified [CE] as desired.  Repeat CERIG command for
    additional rigid region equations.

    This command will generate the constraint equations needed for defining
    rigid lines in 2-D or 3-D space.  Multiple rigid lines relative to a
    common point are used to define a rigid area or a rigid volume.  In 2-D
    space, with Ldof = ALL, three equations are generated for each pair of
    constrained nodes.  These equations define the three rigid body motions
    in global Cartesian space, i.e., two in-plane translations and one in-
    plane rotation.  These equations assume the X-Y plane to be the active
    plane with UX, UY, and ROTZ degrees of freedom available at each node.
    Other types of equations can be generated with the appropriate Ldof
    labels.

    Six equations are generated for each pair of constrained nodes in 3-D
    space (with Ldof = ALL).  These equations define the six rigid body
    motions in global Cartesian space.  These equations assume that UX, UY,
    UZ, ROTX, ROTY, and ROTZ degrees of freedom are available at each node.

    The UXYZ label allows generating a partial set of rigid region
    equations.  This option is useful for transmitting the bending moment
    between elements having different degrees of freedom at a node.  With
    this option only two of the three equations are generated for each pair
    of constrained nodes in 2-D space.  In 3-D space, only three of the six
    equations are generated.  In each case the rotational coupling
    equations are not generated.  Similarly, the RXYZ label allows
    generating a partial set of equations with the translational coupling
    equations omitted.

    Applying this command to a large number of slave nodes may result in
    constraint equations with a large number of coefficients. This may
    significantly increase the peak memory required during the process of
    element assembly. If real memory or virtual memory is not available,
    consider reducing the number of slave nodes.

    Note that under certain circumstances the constraint equations
    generated by CERIG may be modified during the solution. See Program
    Modification of Constraint Equations for more information.

    As an alternative to the CERIG command, you can define a similar type
    of rigid region using contact elements and the internal multipoint
    constraint (MPC) algorithm. See Surface-Based Constraints for more
    information.

    CERIG cannot be deleted using CEDELE,ALL and then regenerated in the
    second or higher load steps if the LSWRITE  and LSSOLVE procedure is
    used. CERIG writes constraint equations directly into load step files.
    Deleting constraint equations (CEDELE,ALL) cannot always maintain the
    consistency among load steps.
    """
    command = f"CERIG,{maste},{slave},{ldof},{ldof2},{ldof3},{ldof4},{ldof5}"
    return self.run(command, **kwargs)


def cesgen(self, itime="", inc="", nset1="", nset2="", ninc="", **kwargs):
    """Generates a set of constraint equations from existing sets.

    APDL Command: CESGEN

    Parameters
    ----------
    itime, inc
        Do this generation operation a total of ITIMEs, incrementing all
        nodes in the existing sets by INC each time after the first.  ITIME
        must be >1 for generation to occur.

    nset1, nset2, ninc
        Generate sets from sets beginning with NSET1 to NSET2 (defaults to
        NSET1) in steps of NINC (defaults to 1).  If NSET1 is negative,
        NSET2 and NINC are ignored and the last |NSET1| sets (in sequence
        from maximum set number) are used as the sets to be repeated.

    Notes
    -----
    Generates additional sets of constraint equations (with same labels)
    from existing sets.  Node numbers between sets may be uniformly
    incremented.
    """
    command = f"CESGEN,{itime},{inc},{nset1},{nset2},{ninc}"
    return self.run(command, **kwargs)


def rbe3(self, master="", dof="", slaves="", wtfact="", **kwargs):
    """Distributes the force/moment applied at the master node to a set  of

    APDL Command: RBE3
    slave nodes, taking into account the geometry of the slave nodes as
    well as weighting factors.

    Parameters
    ----------
    master
        Node at which the force/moment to be distributed will be applied.
        This node must be associated with an element for the master node to
        be included in the DOF solution.

    dof
        Refers to the master node degrees of freedom to be used in
        constraint equations. Valid labels are: UX, UY, UZ, ROTX, ROTY,
        ROTZ, UXYZ, RXYZ, ALL

    slaves
        The name of an array parameter that contains a list of slave nodes.
        Must specify the starting index number. ALL can be used for
        currently selected set of nodes. The slave nodes may not be
        colinear, that is, not be all located on the same straight line
        (see Notes below).

    wtfact
        The name of an array parameter that contains a list of weighting
        factors corresponding to each slave node above. Must have the
        starting index number. If not specified, the weighting factor for
        each slave node defaults to 1.

    Notes
    -----
    The force is distributed to the slave nodes proportional to the
    weighting factors. The moment is distributed as forces to the slaves;
    these forces are proportional to the distance from the center of
    gravity of the slave nodes times the weighting factors.  Only the
    translational degrees of freedom of the slave nodes are used for
    constructing the constraint equations. Constraint equations are
    converted to distributed forces/moments on the slave nodes during
    solution.

    RBE3 creates constraint equations such that the motion of the master is
    the average of the slaves. For the rotations, a least-squares approach
    is used to define the "average rotation" at the master from the
    translations of the slaves. If the slave nodes are colinear, then one
    of the master rotations that is parallel to the colinear direction can
    not be determined in terms of the translations of the slave nodes.
    Therefore, the associated moment component on the master node in that
    direction can not be transmitted. When this case occurs, a warning
    message is issued and the constraint equations created by RBE3 are
    ignored.

    Applying this command to a large number of slave nodes may result in
    constraint equations with a large number of coefficients. This may
    significantly increase the peak memory required during the process of
    element assembly. If real memory or virtual memory is not available,
    consider reducing the number of slave nodes.

    As an alternative to the RBE3 command, you can apply a similar type of
    constraint using contact elements and the internal multipoint
    constraint (MPC) algorithm. See Surface-based Constraints for more
    information.

    This command is also valid in SOLUTION.
    """
    command = f"RBE3,{master},{dof},{slaves},{wtfact}"
    return self.run(command, **kwargs)



def fc(self, mat="", lab1="", lab2="", data1="", data2="", data3="",
       data4="", data5="", data6="", **kwargs):
    """Provides failure criteria information and activates a data

    APDL Command: FC
    table to input temperature-dependent stress and strain limits.

    Parameters
    ----------
    mat
        Material reference number. You can define failure criteria
        for up to 250 different materials.

    lab1
        Type of data.

        TEMP - Temperatures. Each of the materials you define can
        have a different set of temperatures to define the failure
        criteria.

        EPEL - Strains.

        S - Stresses.

    lab2
        Specific criteria. Not used if Lab1 = TEMP.

        XTEN - Allowable tensile stress or strain in the
        x-direction. (Must be positive.)

        XCMP - Allowable compressive stress or strain in the
        x-direction. (Defaults to negative of XTEN.)

        YTEN - Allowable tensile stress or strain in the
        y-direction. (Must be positive.)

        YCMP - Allowable compressive stress or strain in the
        y-direction. (Defaults to negative of YTEN.)

        ZTEN - Allowable tensile stress or strain in the
        z-direction. (Must be positive.)

        ZCMP - Allowable compressive stress or strain in the
        z-direction. (Defaults to negative of ZTEN.)

        XY - Allowable XY stress or shear strain. (Must be
        positive.)

        YZ - Allowable YZ stress or shear strain. (Must be
        positive.)

        XZ - Allowable XZ stress or shear strain. (Must be
        positive.)

        XYCP - XY coupling coefficient (Used only if Lab1 =
        S). Defaults to -1.0. [1]

        YZCP - YZ coupling coefficient (Used only if Lab1 =
        S). Defaults to -1.0. [1]

        XZCP - XZ coupling coefficient (Used only if Lab1 =
        S). Defaults to -1.0. [1]

    data1, data2, data3, . . . , data6
        Description of DATA1 through DATA6.

        T1, T2, T3, T4, T5, T6 - Temperature at which limit data
        is input. Used only when Lab1 = TEMP.

        V1, V2, V3, V4, V5, V6 - Value of limit stress or strain
        at temperature T1 through T6. Used only when Lab1 = S or
        EPEL.

    Notes
    -----
    The data table can be input in either PREP7 or POST1. This
    table is used only in POST1. When you postprocess failure
    criteria results defined via the FC command (PLESOL, PRESOL,
    PLNSOL, PRNSOL, PRRSOL, etc.), the active coordinate system
    must be the coordinate system of the material being
    analyzed. You do this using RSYS, SOLU. For layered
    applications, you also use the LAYER command. See the specific
    element documentation in the Element Reference for information
    on defining your coordinate system for layers.

    Some plotting and printing functions will not support Failure
    Criteria for your PowerGraphics displays. This could result in
    minor changes to other data when Failure Criteria are
    applied. See the appropriate plot or print command
    documentation for more information.
    """
    command = "FC,%s,%s,%s,%s,%s,%s,%s,%s,%s" % (str(mat), str(lab1), str(
        lab2), str(data1), str(data2), str(data3), str(data4), str(data5), str(data6))
    return self.run(command, **kwargs)


def fccheck(self, **kwargs):
    """Checks both the strain and stress input criteria for all materials.

     APDL Command: FCCHECK

    Notes
    -----
    Issue the FCCHECK command to check the completeness of the input during
    the input phase.
    """
    command = "FCCHECK,"
    return self.run(command, **kwargs)


def fcdele(self, mat="", **kwargs):
    """Deletes previously defined failure criterion data for the given

    APDL Command: FCDELE
    material.

    Parameters
    ----------
    mat
        Material number. Deletes all FC command input for this material.

    Notes
    -----
    This command is also valid in POST1.
    """
    command = "FCDELE,%s" % (str(mat))
    return self.run(command, **kwargs)


def fclist(self, mat="", temp="", **kwargs):
    """To list what the failure criteria is that you have input.

    APDL Command: FCLIST

    Parameters
    ----------
    mat
         Material number (defaults to ALL for all materials).

    temp
        Temperature to be evaluated at (defaults to TUNIF).

    Notes
    -----
    This command allows you to see what you have already input for failure
    criteria using the FC commands.
    """
    return self.run(f"FCLIST,{mat},,{temp}", **kwargs)


def fctyp(self, oper="", lab="", **kwargs):
    """Activates or removes failure-criteria types for postprocessing.

    APDL Command: FCTYP

    Parameters
    ----------
    oper
        Operation key:

        ADD - Activate failure-criteria types. This option is the default behavior.

        DELE - Remove failure-criteria types.

    lab
        Valid failure-criteria labels. If ALL, select all available
        (including user-defined) failure criteria.

        EMAX - Maximum strain criterion (default)

        SMAX - Maximum stress criterion (default)

        TWSI  - Tsai-Wu strength index (default)

        TWSR  - Inverse of Tsai-Wu strength ratio index (default)

        HFIB  - Hashin fiber failure criterion

        HMAT  - Hashin matrix failure criterion

        PFIB  - Puck fiber failure criterion

        PMAT  - Puck inter-fiber (matrix) failure criterion

        L3FB - LaRc03 fiber failure criterion

        L3MT - LaRc03 matrix failure criterion

        L4FB - LaRc04 fiber failure criterion

        L4MT - LaRc04 matrix failure criterion

        USR1 through USR9  - User-defined failure criteria

    Notes
    -----
    The FCTYP command modifies the list of active failure criteria.

    By default, active failure criteria include EMAX, SMAX, TWSI, and TWSR.

    The command affects any subsequent postprocessing listing and plotting
    commands (such as PRESOL, PRNSOL, PLESOL, PLNSOL, and ETABLE).

    A single FCTYP command allows up to six failure-criteria labels. If
    needed, reissue the command to activate or remove additional failure-
    criteria types.
    """
    command = f"FCTYP,{oper},{lab}"
    return self.run(command, **kwargs)
